from typing import Dict, Union

from sfaira.versions.metadata import OntologyList, OntologyCl
from sfaira.versions.metadata import OntologyCellosaurus, OntologyHancestro, OntologyHsapdv, OntologyMondo, \
    OntologyMmusdv, OntologyEfo, OntologySex, OntologyTaxon, OntologyUberon, OntologyUberonLifecyclestage

OTHER_ORGANISM_KEY = "other"

DEFAULT_CL = "v2021-08-10"
DEFAULT_HSAPDV = "master"
DEFAULT_MONDO = "v2021-08-11"
DEFAULT_MMUSDV = "master"
DEFAULT_PATO = "v2021-08-06"
DEFAULT_NCBITAXON = "v2021-06-10"
DEFAULT_UBERON = "v2021-07-27"


class OntologyContainerSfaira:

    """
    The attributes that are relayed via properties, which have a corresponding private attribute "_*", are used to
    lazily load these ontologies upon usage and redistribute loading time from package initialisation to actual
    usage of ontology.
    """

    _assay_sc: Union[None, OntologyEfo]
    _cell_line: Union[None, OntologyCellosaurus]
    _cell_type: Union[None, OntologyCl]
    _development_stage: Union[None, Dict[str, Union[OntologyHsapdv, OntologyMmusdv]]]
    _ethnicity: Union[None, Dict[str, Union[OntologyHancestro, None]]]
    _organ: Union[None, OntologyUberon]
    _organism: Union[None, OntologyTaxon]
    _sex: Union[None, OntologySex]

    def __init__(self):
        self.key_other = OTHER_ORGANISM_KEY

        self.annotated = OntologyList(terms=[True, False])
        self.author = None
        self.assay_differentiation = None
        self._assay_sc = None
        self.assay_type_differentiation = OntologyList(terms=["guided", "unguided"])
        self.bio_sample = None
        self._cell_line = None
        self._cell_type = None
        self.collection_id = None
        self.default_embedding = None
        self._development_stage = None
        self._disease = None
        self.doi = None
        self.doi_main = None
        self.doi_journal = None
        self.doi_preprint = None
        self._ethnicity = None
        self.feature_type = OntologyList(terms=["rna", "peak", "protein"])
        self.id = None
        self.individual = None
        self.normalization = None
        self._organ = None
        self._organism = None
        self.primary_data = OntologyList(terms=[True, False])
        self.sample_source = OntologyList(terms=["primary_tissue", "2d_culture", "3d_culture", "tumor"])
        self._sex = None
        self.supplier = OntologyList(terms=["cellxgene", "sfaira"])
        self.tech_sample = None
        self.title = None
        self.year = OntologyList(terms=list(range(2000, 3000)))

    def reload_ontology(self, attr):
        """
        Complex alternative to attribute-wise setters.

        :param attr:
        :return:
        """
        kwargs = {"recache": True}
        if attr == "assay_sc":
            self._assay_sc = OntologyEfo(**kwargs)
        elif attr == "cell_line":
            self._cell_line = OntologyCellosaurus(**kwargs)
        elif attr == "cell_type":
            self._cell_type = OntologyCl(branch=DEFAULT_CL, **kwargs)
        elif attr == "development_stage":
            self._development_stage = {
                "Homo sapiens": OntologyHsapdv(branch=DEFAULT_HSAPDV, **kwargs),
                "Mus musculus": OntologyMmusdv(branch=DEFAULT_MMUSDV, **kwargs),
                self.key_other: OntologyUberonLifecyclestage(branch=DEFAULT_UBERON, **kwargs),
            }
        elif attr == "disease":
            self._disease = OntologyMondo(branch=DEFAULT_MONDO, **kwargs)
        elif attr == "ethnicity":
            self._ethnicity = {
                "homosapiens": OntologyHancestro(),
                self.key_other: None,
            }
        elif attr == "organ":
            self._organ = OntologyUberon(branch=DEFAULT_UBERON, **kwargs)
        elif attr == "organism":
            self._organism = OntologyTaxon(branch=DEFAULT_NCBITAXON, **kwargs)
        elif attr == "sex":
            self._sex = OntologySex(branch=DEFAULT_PATO, **kwargs)

    @property
    def assay_sc(self):
        if self._assay_sc is None:  # Lazy loading after class instantiation.
            self._assay_sc = OntologyEfo()
        return self._assay_sc

    @property
    def cell_line(self):
        if self._cell_line is None:  # Lazy loading after class instantiation.
            self._cell_line = OntologyCellosaurus()
        return self._cell_line

    @property
    def cell_type(self):
        if self._cell_type is None:  # Lazy loading after class instantiation.
            self._cell_type = OntologyCl(branch=DEFAULT_CL)
        return self._cell_type

    @cell_type.setter
    def cell_type(self, x: str):
        self._cell_type = OntologyCl(branch=x)

    @property
    def development_stage(self):
        if self._development_stage is None:  # Lazy loading after class instantiation.
            self._development_stage = {
                "Homo sapiens": OntologyHsapdv(branch=DEFAULT_HSAPDV),
                "Mus musculus": OntologyMmusdv(branch=DEFAULT_MMUSDV),
                self.key_other: OntologyUberonLifecyclestage(branch=DEFAULT_UBERON),
            }
        return self._development_stage

    @property
    def disease(self):
        if self._disease is None:  # Lazy loading after class instantiation.
            self._disease = OntologyMondo(branch=DEFAULT_MONDO)
        return self._disease

    @property
    def ethnicity(self):
        if self._ethnicity is None:  # Lazy loading after class instantiation.
            self._ethnicity = {
                "Homo sapiens": OntologyHancestro(),
                self.key_other: None,
            }
        return self._ethnicity

    @property
    def organ(self):
        if self._organ is None:  # Lazy loading after class instantiation.
            self._organ = OntologyUberon(branch=DEFAULT_UBERON)
        return self._organ

    @property
    def organism(self):
        if self._organism is None:  # Lazy loading after class instantiation.
            self._organism = OntologyTaxon(branch=DEFAULT_NCBITAXON)
        return self._organism

    @property
    def sex(self):
        if self._sex is None:  # Lazy loading after class instantiation.
            self._sex = OntologySex(branch=DEFAULT_PATO)
        return self._sex



import os

from setuptools import setup, find_packages
import versioneer

author = 'theislab'
author_email = 'david.fischer@helmholtz-muenchen.de'
description = "sfaira is a model and a data repository for single-cell data in a single python package."

with open("README.rst", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()


def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


WD = os.path.dirname(__file__)
templates = package_files(os.path.join(WD, "sfaira", "commands", "templates"))

setup(
    name='sfaira',
    author=author,
    author_email=author_email,
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    packages=find_packages(include=['sfaira', 'sfaira.*']),
    include_package_data=True,
    package_data={'': templates},
    entry_points={
        'console_scripts': [
            'sfaira=sfaira.cli:main',
        ],
    },
    install_requires=requirements,
    extras_require={
        'tensorflow': [
            # 'tensorflow>=2.0.0'  # TODO Add Tensorflow here again
        ],
        'plotting_deps': [
            "seaborn",
            "matplotlib",
            "sklearn"
        ],
        'extension': [
            "sfaira_extension",
        ],
    },
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
)



from sfaira.data.store.io.load_store import load_store
from sfaira.data.store.stores import StoreSingleFeatureSpace, \
    StoreAnndata, StoreDao, \
    StoreMultipleFeatureSpaceBase, \
    StoresAnndata, StoresDao, StoresH5ad



import numpy as np
import os
from typing import Union
import pandas as pd
import urllib.request

from sfaira import settings
from sfaira.ui import UserInterface
from sfaira.unit_tests.data_for_tests.loaders.utils import PrepareData
from sfaira.unit_tests import DIR_TEMP


class HelperUi:
    ui: Union[UserInterface]
    data: np.ndarray

    """
    Contains functions _test* to test individual functions and attributes of the user ui class.

    TODO for everybody working on this, add one _test* function in here and add it into
    basic_estimator_test(). See _test_call() for an example.
    """

    def __init__(self):
        self.temp_fn = os.path.join(DIR_TEMP, "test_data")

    def prepare_local_tempfiles(self):
        # create temp directory
        if not os.path.exists(self.temp_fn):
            os.makedirs(self.temp_fn)
        # download an example weight from sfaira repo
        lookuptable = pd.read_csv(
            os.path.join(settings.sfaira_repo_url, 'model_lookuptable.csv'),
            header=0,
            index_col=0
        )
        url = lookuptable.loc[0, "model_file_path"]
        if os.path.basename(url) not in os.listdir(self.temp_fn):
            urllib.request.urlretrieve(url, os.path.join(self.temp_fn, os.path.basename(url)))

    def _get_adata(self):
        """
        Create an adata object for use in unit tests

        :return:
        """
        dsg = PrepareData().prepare_dsg(rewrite=True, load=False)
        dsg.subset(key="id", values=["homosapiens_lung_2021_None_mock4_001_no_doi_mock4"])
        dsg.load()
        return dsg.adata

    def test_local_repo_ui_init(self):
        """
        Test if the sfaira UI can be sucessfully initialised using a local model repository

        :return:
        """
        self.ui = UserInterface(custom_repo=self.temp_fn, sfaira_repo=False)

    def test_public_repo_ui_init(self):
        """
        Test if the sfaira UI can be sucessfully initialised using the public sfaira model repository

        :return:
        """
        self.ui = UserInterface(custom_repo=None, sfaira_repo=True, cache_path=self.temp_fn)

    def test_data_and_model_loading(self):
        self.ui = UserInterface(custom_repo=None, sfaira_repo=True, cache_path=self.temp_fn)
        self.ui.zoo_embedding.model_id = 'embedding_human-blood-ae-0.2-0.1_theislab'
        self.ui.zoo_celltype.model_id = 'celltype_human-blood-mlp-0.1.3-0.1_theislab'
        test_data = self._get_adata()
        self.ui.load_data(test_data, gene_ens_col='index')
        self.ui.load_model_celltype()
        self.ui.load_model_embedding()
        self.ui.predict_all()
        assert "X_sfaira" in self.ui.data.adata.obsm_keys()
        assert "celltypes_sfaira" in self.ui.data.adata.obs_keys()


def test_ui():
    ui = HelperUi()
    ui.prepare_local_tempfiles()
    ui.test_public_repo_ui_init()
    ui.test_local_repo_ui_init()
    ui.test_data_and_model_loading()



import anndata
import os
import pandas as pd

from sfaira.data import DatasetBase


class Dataset(DatasetBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE130nnn/GSE130148/suppl/GSE130148%5Fraw%5Fcounts%2Ecsv%2Egz"
        self.download_url_meta = "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE130nnn/GSE130148/suppl/GSE130148%5Fbarcodes%5Fcell%5Ftypes%2Etxt%2Egz"

        self.assay_sc = "Drop-seq"
        self.author = "Braga"
        self.disease = "healthy"
        self.doi_journal = "10.1038/s41591-019-0468-5"
        self.layer_counts = "X"
        self.organ = "lung"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.sample_source = "primary_tissue"
        self.state_exact = "uninvolved areas of tumour resection material"
        self.year = 2019

        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"
        self.cell_type_obs_key = "celltype"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    fn = [
        os.path.join(data_dir, "GSE130148_raw_counts.csv.gz"),
        os.path.join(data_dir, "GSE130148_barcodes_cell_types.txt.gz"),
    ]
    adata = anndata.read_csv(fn[0]).T
    adata.obs = pd.read_csv(fn[1], sep="\t", index_col=0)

    return adata



import os
import tarfile
import gzip
from io import StringIO
import anndata as ad
import pandas as pd
import scipy.sparse

from sfaira.data import DatasetBase


class Dataset(DatasetBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE81nnn/GSE81547/suppl/GSE81547_RAW.tar"
        self.download_url_meta = "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE81nnn/GSE81547/matrix/GSE81547_series_matrix.txt.gz"

        self.author = "Enge"
        self.disease = "healthy"
        self.doi_journal = "10.1016/j.cell.2017.09.004"
        self.doi_preprint = "10.1101/108043"
        self.layer_counts = "X"
        self.assay_sc = "Smart-seq2"
        self.organ = "islet of Langerhans"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.year = 2017
        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"
        self.cell_type_obs_key = "celltype"
        self.sample_source = "primary_tissue"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    fn = [
        os.path.join(data_dir, "GSE81547_RAW.tar"),
        os.path.join(data_dir, "GSE81547_series_matrix.txt.gz")
    ]
    dfs = []
    with tarfile.open(fn[0]) as tar:
        for member in tar.getmembers():
            d = pd.read_csv(tar.extractfile(member), compression="gzip", header=None, sep="\t", index_col=0,
                            names=[member.name.split("_")[0]])
            dfs.append(d)
    adata = ad.AnnData(pd.concat(dfs, axis=1).iloc[1:-6].T)
    adata.X = scipy.sparse.csc_matrix(adata.X)
    with gzip.open(fn[1]) as f:
        file_content = [i.decode("utf-8") for i in f.readlines()]
    inputstring = ""
    for line in file_content:
        if "ID_REF" in line:
            inputstring += line
        if "!Sample_title" in line:
            inputstring += line[1:]
        if "!Sample_characteristics_ch1\t\"inferred_cell_type: alpha" in line:
            inputstring += line[1:]
    data = StringIO(inputstring)
    d = pd.read_csv(data, sep="\t").T
    d.columns = d.iloc[0]
    d.drop("Sample_title", inplace=True)
    d = d.reset_index().set_index("ID_REF")
    d.columns.name = None
    d.index.name = None
    adata.obs["celltype"] = [d.loc[i]["Sample_characteristics_ch1"].split(": ")[1] for i in adata.obs.index]
    adata.obs["patient"] = ["_".join(d.loc[i]["index"].split("_")[:2]) for i in adata.obs.index]

    return adata



import os
import pandas as pd
import scanpy as sc

# This is provided in plain text on GEO
sample_map = {"Sample1": "nCoV 1 scRNA-seq",
              "Sample2": "nCoV 2 scRNA-seq",
              "Sample3": "Flu 1 scRNA-seq",
              "Sample4": "Flu 2 scRNA-seq",
              "Sample5": "Normal 1 scRNA-seq",
              "Sample6": "Flu 3 scRNA-seq",
              "Sample7": "Flu 4 scRNA-seq",
              "Sample8": "Flu 5 scRNA-seq",
              "Sample9": "nCoV 3 scRNA-seq",
              "Sample10": "nCoV 4 scRNA-seq",
              "Sample11": "nCoV 5 scRNA-seq",
              "Sample12": "nCoV 6 scRNA-seq",
              "Sample13": "Normal 2 scRNA-seq",
              "Sample14": "Normal 3 scRNA-seq",
              "Sample15": "nCoV 7 scRNA-seq",
              "Sample16": "nCoV 8 scRNA-seq",
              "Sample17": "nCoV 9 scRNA-seq",
              "Sample18": "nCoV 10 scRNA-seq",
              "Sample19": "Normal 4 scRNA-seq",
              "Sample20": "nCoV 11 scRNA-seq"
              }


def load(data_dir, sample_fn, **kwargs):
    adata = sc.read_10x_mtx(data_dir, prefix="GSE149689_")
    adata.obs["sample"] = "Sample" + adata.obs.index.str.split("-").str[1]
    adata.obs["GEO upload info"] = adata.obs["sample"].map(sample_map)

    fn_meta = os.path.join(data_dir, "Table_S1.xlsx")
    metadata = pd.read_excel(fn_meta, engine="openpyxl")
    metadata.fillna(method="ffill", inplace=True)
    metadata.replace(to_replace="(\\n)", value=" ", regex=True, inplace=True)
    metadata.rename(columns={"Experimental\nbatch": "Experimental batch"}, inplace=True)
    metadata.drop_duplicates(inplace=True)

    adata.obs = adata.obs.reset_index().merge(metadata, how="left").set_index("index")
    adata.obs["Age"] = [str(x) for x in adata.obs["Age"].values]
    return adata



import anndata
import os
import pandas as pd

from sfaira.data import DatasetBase


class Dataset(DatasetBase):

    """
    TODO extra meta data in obs2

    age: columns "Age" contains integer entries and Unknown
    diseases: column "Diagnosis" contains entries NSIP, cHP, Control, IPF, ILD, Sarcoidosis
        column Tobacco contains entries Y,N
    ethnicity: column "Ethnicity" contains entries African_American, Caucasian, Hispanic, Unknown
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = [
            "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE135nnn/GSE135893/suppl/GSE135893%5Fmatrix%2Emtx%2Egz",
            "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE135nnn/GSE135893/suppl/GSE135893%5Fgenes%2Etsv%2Egz",
            "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE135nnn/GSE135893/suppl/GSE135893%5Fbarcodes%2Etsv%2Egz"
        ]
        self.download_url_meta = [
            "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE135nnn/GSE135893/suppl/GSE135893%5FIPF%5Fmetadata%2Ecsv%2Egz",
            "https://advances.sciencemag.org/highwire/filestream/234522/field_highwire_adjunct_files/2/aba1972_Table_S2.csv",
        ]

        self.author = "Habermann"
        self.doi_journal = "10.1126/sciadv.aba1972"
        self.doi_preprint = "10.1101/753806"
        self.layer_counts = "X"
        self.organ = "lung parenchyma"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.assay_sc_obs_key = "Chemistry"
        self.year = 2020
        self.sample_source = "primary_tissue"
        self.sex_obs_key = "Gender"
        self.tech_sample_obs_key = "Sample_Name"

        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"

        self.cell_type_obs_key = "celltype"
        self.state_exact_obs_key = "Diagnosis"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    fn = [
        os.path.join(data_dir, "GSE135893_matrix.mtx.gz"),
        os.path.join(data_dir, "GSE135893_genes.tsv.gz"),
        os.path.join(data_dir, "GSE135893_barcodes.tsv.gz"),
        os.path.join(data_dir, "GSE135893_IPF_metadata.csv.gz"),
        os.path.join(data_dir, "aba1972_Table_S2.csv"),
    ]
    adata = anndata.read_mtx(fn[0]).T
    adata.var = pd.read_csv(fn[1], index_col=0, header=None, names=["ids"])
    adata.obs = pd.read_csv(fn[2], index_col=0, header=None, names=["barcodes"])
    obs = pd.read_csv(fn[3], index_col=0)
    obs2 = pd.read_csv(fn[4], index_col=0)
    obs["Chemistry"] = [{"3_prime_V2": "10x 3' v2", "5_prime": "10x 5' v1"}[obs2.loc[x, "Chemistry"]]
                        for x in obs["orig.ident"].values]
    obs["Gender"] = [{"F": "female", "M": "male", "Unknown": "unknown"}[obs2.loc[x, "Gender"]]
                     for x in obs["orig.ident"].values]
    adata = adata[obs.index.tolist(), :].copy()
    adata.obs = obs

    return adata



VAEIAF_TOPOLOGIES = {
    "0.1": {
        "model_type": "vaeiaf",
        "input": {
            "genome": "Homo_sapiens.GRCh38.102",
            "genes": ["biotype", "protein_coding"],
        },
        "output": {},
        "hyper_parameters": {
            "latent_dim": (256, 128, 64, 128, 256),
            "n_iaf": 2,
            "l1_coef": 0.,
            "l2_coef": 0.,
            "dropout_rate": 0.,
            "batchnorm": True,
            "activation": "tanh",
            "init": "glorot_uniform",
            "output_layer": "nb_shared_disp"
        }
    },
    "0.2": {
        "model_type": "vaeiaf",
        "input": {
            "genome": "Homo_sapiens.GRCh38.102",
            "genes": ["biotype", "protein_coding"],
        },
        "output": {},
        "hyper_parameters": {
            "latent_dim": (512, 256, 128, 256, 512),
            "n_iaf": 2,
            "l1_coef": 0.,
            "l2_coef": 0.,
            "dropout_rate": 0.,
            "batchnorm": True,
            "activation": "tanh",
            "init": "glorot_uniform",
            "output_layer": "nb_shared_disp"
        }
    }
}

# Load versions from extension if available:
try:
    from sfaira_extension.versions.topology_versions.human.embedding import VAEIAF_TOPOLOGIES as VAEIAF_TOPOLOGIES_EXTENSION
    VAEIAF_TOPOLOGIES = {
        **VAEIAF_TOPOLOGIES,
        **VAEIAF_TOPOLOGIES_EXTENSION
    }
except ImportError:
    pass



import numpy as np
import pandas
import pickle
import json
import shutil
import warnings
from typing import Union, List
import os

from sfaira.consts import OCS
from sfaira.data import load_store
from sfaira.data.dataloaders import Universe
from sfaira.estimators.keras.base import EstimatorKerasEmbedding
from sfaira.ui import ModelZoo
from sfaira.versions.metadata import CelltypeUniverse, OntologyCl


def _tp(yhat, ytrue):
    """
    Class wise true positive count.

    :param yhat:
    :param ytrue:
    :return:
    """
    yhat_true = np.asarray(yhat == np.max(yhat, axis=1, keepdims=True), dtype="float32")
    return np.sum(yhat_true * ytrue, axis=0)


def _fp(yhat, ytrue):
    """
    Class wise false positive count.

    :param yhat:
    :param ytrue:
    :return:
    """
    yhat_true = np.asarray(yhat == np.max(yhat, axis=1, keepdims=True), dtype="float32")
    return np.sum(yhat_true * (1. - ytrue), axis=0)


def _tn(yhat, ytrue):
    """
    Class wise true negative count.

    :param yhat:
    :param ytrue:
    :return:
    """
    yhat_true = np.asarray(yhat < np.max(yhat, axis=1, keepdims=True), dtype="float32")
    return np.sum(yhat_true * (1. - ytrue), axis=0)


def _fn(yhat, ytrue):
    """
    Class wise false negative count.

    :param yhat:
    :param ytrue:
    :return:
    """
    yhat_true = np.asarray(yhat < np.max(yhat, axis=1, keepdims=True), dtype="float32")
    return np.sum(yhat_true * ytrue, axis=0)


def accuracy(yhat, ytrue):
    """
    Class wise accuracy.

    :param yhat:
    :param ytrue:
    :return:
    """
    return (_tp(yhat, ytrue) + _tn(yhat, ytrue)) / yhat.shape[0]


def f1(yhat, ytrue):
    """
    Class wise F1.

    :param yhat:
    :param ytrue:
    :return:
    """
    precision = _tp(yhat, ytrue) / (_tp(yhat, ytrue) + _fp(yhat, ytrue))
    recall = _tp(yhat, ytrue) / (_tp(yhat, ytrue) + _fn(yhat, ytrue))
    return 2 * 1 / (1 / precision + 1 / recall)


def auc_roc(yhat, ytrue):
    """
    Class wise AUC ROC.

    :param yhat:
    :param ytrue:
    :return:
    """
    import sklearn

    auc_roc = np.array([
        sklearn.metrics.roc_auc_score(y_true=ytrue[:, i], y_score=yhat[:, i])
        for i in range(ytrue.shape[0])
    ])
    return auc_roc


class GridsearchContainer:

    histories: Union[None, dict]
    evals: Union[None, dict]
    run_ids: Union[None, list]
    gs_keys: Union[None, dict]
    summary_tab: Union[None, pandas.DataFrame]
    cv: bool
    source_path: dict
    model_id_len: Union[None, int]

    def __init__(
            self,
            source_path: dict,
            cv: bool
    ):
        self.histories = None
        self.evals = None
        self.run_ids = None
        self.gs_keys = None
        self.cv = cv
        self.source_path = source_path
        self.summary_tab = None

    def load_gs(
            self,
            gs_ids: List[str]
    ):
        """
        Loads all relevant data of a grid search.

        :param gs_ids:
        :return:
        """
        res_dirs = [
            os.path.join(self.source_path[x], x, "results", "")
            for x in gs_ids
        ]
        run_ids = [
            np.unique([
                x.split("_evaluation.pickle")[0]
                for x in os.listdir(indir)
                if "_evaluation.pickle" in x
            ])
            for i, indir in enumerate(res_dirs)
        ]
        histories = {}
        evals = {}
        hyperpars = {}
        model_hyperpars = {}
        run_ids_proc = []
        gs_keys = []
        gs_dirs = []
        for i, indir in enumerate(res_dirs):
            for x in run_ids[i]:
                fn_history = os.path.join(indir, f"{x}_history.pickle")
                if os.path.isfile(fn_history):
                    with open(fn_history, 'rb') as f:
                        histories[x] = pickle.load(f)
                else:
                    print(f"file {x}_history.pickle not found")

                fn_eval = os.path.join(indir, f"{x}_evaluation.pickle")
                if os.path.isfile(fn_eval):
                    with open(fn_eval, 'rb') as f:
                        evals[x] = pickle.load(f)
                else:
                    print(f"file {x}_evaluation.pickle not found")

                fn_hp = os.path.join(indir, f"{x}_hyperparam.pickle")
                if os.path.isfile(fn_hp):
                    with open(fn_hp, 'rb') as f:
                        hyperpars[x] = pickle.load(f)
                else:
                    print(f"file {x}_hyperparam.pickle not found")

                fn_mhp = os.path.join(indir, f"{x}_model_hyperparam.pickle")
                if os.path.isfile(fn_mhp):
                    with open(fn_mhp, 'rb') as f:
                        model_hyperpars[x] = pickle.load(f)
                else:
                    print(f"file {x}_model_hyperparam.pickle not found")

                run_ids_proc.append(x)
                gs_keys.append(os.path.normpath(indir).split(os.path.sep)[-2])
                gs_dirs.append(indir)

        self.run_ids = run_ids_proc
        self.gs_keys = dict(zip(run_ids_proc, gs_keys))
        self.gs_dirs = dict(zip(run_ids_proc, gs_dirs))
        self.evals = evals
        self.hyperpars = hyperpars
        self.model_hyperpars = model_hyperpars
        self.histories = histories

    def load_y(
            self,
            hat_or_true: str,
            run_id: str
    ):
        fn = os.path.join(self.gs_dirs[run_id], f"{run_id}_y{hat_or_true}.npy")
        return np.load(fn)

    def best_model_by_partition(
            self,
            partition_select: str,
            metric_select: str,
            cv_mode: str = "mean",
            subset: dict = {},
            return_run_only: bool = False,
            grouping: list = ["organ", "model_type"]
    ):
        """

        :param partition_select:
        :param metric_select:
        :param cv_mode:
        :param subset:
        :param return_run_only:
        :param grouping:
        :return:
        """
        model_ids = []
        run_ids = []
        for id, df in self.summary_tab.groupby(grouping):
            if df.shape[0] > 0:
                model_id_temp, run_id_temp, cv_id_temp = self.get_best_model_ids(
                    tab=df,
                    partition_select=partition_select,
                    metric_select=metric_select,
                    subset=subset,
                    cv_mode=cv_mode
                )
                model_ids.append(model_id_temp)
                run_ids.append(run_id_temp)
        if return_run_only:
            return self.summary_tab.loc[[x in run_ids for x in self.summary_tab["run"].values], :]
        else:
            return self.summary_tab.loc[[x in model_ids for x in self.summary_tab["model_gs_id"].values], :]

    def get_best_model_ids(
            self,
            tab,
            metric_select: str,
            partition_select: str,
            subset: dict = {},
            cv_mode: str = "mean"
    ):
        """

        :param tab:
        :param metric_select:
        :param partition_select:
        :param subset:
        :param cv_mode:
        :return:
        """
        for k, v in subset.items():
            tab = tab.loc[tab[k].values == v, :]

        if metric_select.endswith('accuracy') \
                or metric_select.endswith('acc_agg') \
                or metric_select.endswith('f1') \
                or metric_select.endswith('tpr'):
            ascending = False
            if cv_mode == "min":
                raise Warning("selected cv_mode min with metric_id %s, likely not intended" % metric_select)
        elif metric_select.endswith('loss') \
                or metric_select.endswith('mse') \
                or metric_select.endswith('negll') \
                or metric_select.endswith('custom_cce_agg') \
                or metric_select.endswith('fpr'):
            ascending = True
            if cv_mode == "max":
                raise Warning("selected cv_mode max with metric_id %s, likely not intended" % metric_select)
        else:
            raise ValueError("measure %s not recognized" % metric_select)

        if partition_select not in ["test", "val", "train"]:
            raise ValueError("partition %s not recognised" % partition_select)

        metric_select = f"{partition_select}_{metric_select}"

        if cv_mode.lower() == "mean":
            best_model = tab.groupby("run", as_index=False)[metric_select].mean(). \
                sort_values([metric_select], ascending=ascending)
        elif cv_mode.lower() == "median":
            best_model = tab.groupby("run", as_index=False)[metric_select].median(). \
                sort_values([metric_select], ascending=ascending)
        elif cv_mode.lower() == "max":
            best_model = tab.groupby("run", as_index=False)[metric_select].max(). \
                sort_values([metric_select], ascending=ascending)
        elif cv_mode.lower() == "min":
            best_model = tab.groupby("run", as_index=False)[metric_select].min(). \
                sort_values([metric_select], ascending=ascending)
        else:
            raise ValueError("cv_mode %s not recognized" % cv_mode)

        best_run_id = best_model['run'].values[0] if best_model.shape[0] > 0 else None

        best_cv = tab[tab["run"] == best_run_id]. \
            sort_values([metric_select], ascending=ascending)['cv'].values[0] if best_run_id is not None \
            else None

        best_model_id = tab[tab["run"] == best_run_id]. \
            sort_values([metric_select], ascending=ascending)['model_gs_id'].values[0] if best_run_id is not None \
            else None

        return best_model_id, best_run_id, best_cv

    @property
    def cv_keys(self) -> List[str]:
        """
        Returns keys of cross-validation used in dictionaries in this class.

        :return: list of string keys
        """
        return np.unique(self.summary_tab["cv"].values).tolist()

    def save_best_weight(
            self,
            path: str,
            partition: str = "val",
            metric: str = "loss",
            subset: dict = {}
    ):
        """
        Copies weight file from best hyperparameter setting from grid search directory to zoo directory with cleaned
        file name.

        :param path: Target file to save to. This is intended to be the zoo directory ready for upload.
        :param partition:
        :param metric:
        :param subset:
        :return:
        """
        assert not self.cv, "not implemented for CV yet"
        model_id, _, _ = self.get_best_model_ids(
            tab=self.summary_tab,
            partition_select=partition,
            metric_select=metric,
            cv_mode="mean",
            subset=subset,
        )
        shutil.copyfile(
            os.path.join(self.gs_dirs[model_id], self.gs_keys[model_id], "results", f"{model_id}_weights.h5"),
            os.path.join(path, f"{model_id}_weights.h5")
        )

    def plot_completions(
            self,
            groupby=["depth", "width", "lr", "dropout", "l1", "l2"],
            height_fig=7,
            width_fig=7
    ):
        """
        Plot number of completed grid search points by category.

        :param groupby:
        :param height_fig:
        :param width_fig:
        :return:
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        if self.summary_tab is None:
            self.create_summary_tab()
        sns_tab = self.summary_tab.copy()

        # Build figure.
        organs = np.unique(sns_tab["organ"].values)
        model_types = np.unique(sns_tab["model_type"].values)
        hm = np.zeros((len(organs), len(model_types)))
        for i, m in enumerate(model_types):
            for j, o in enumerate(organs):
                n_by_gridpoint = sns_tab.loc[
                    np.logical_and(
                        sns_tab["model_type"].values == m,
                        sns_tab["organ"].values == o
                    ), :].groupby(groupby).size().values
                # Assume that largest number of successful completions is maximum (all completed:
                hm[j, i] = np.sum(n_by_gridpoint == np.max(n_by_gridpoint)) if len(n_by_gridpoint) > 0 else 0
        sns_data_heatmap = pandas.DataFrame(
            hm, index=organs, columns=model_types
        )
        fig, axs = plt.subplots(1, 1, figsize=(height_fig, width_fig))
        with sns.axes_style("dark"):
            axs = sns.heatmap(
                sns_data_heatmap,
                annot=True, fmt=".2f",
                ax=axs,
                xticklabels=True, yticklabels=True,
                cbar_kws={'label': 'n'}
            )
        return fig, axs

    def plot_best_model_by_hyperparam(
            self,
            metric_select: str,
            param_hue='lr',
            partition_select: str = "val",
            partition_show: str = "test",
            subset: dict = {},
            param_x: Union[tuple, list] = ('lr', 'depth', 'width', 'dropout', 'l1', 'l2'),
            show_swarm: bool = False,
            panel_width: float = 4.,
            panel_height: float = 2.
    ):
        """
        Produces boxplots for all hyperparameters choices by organ.

        :param partition: "train" or "eval" or "test" partition of data.
        :param metric_select: Metric to plot.
        :param param_x: Hyper-parameter for x-axis partition.
        :param param_hue: Hyper-parameter for hue-axis partition.
        :param panel_width:
        :param panel_height:
        :return:
        """
        import seaborn as sns
        import matplotlib.pyplot as plt

        params = [param for param in param_x if len(np.unique(self.summary_tab[param])) > 1 and param != param_hue]

        organs = np.unique(self.summary_tab["organ"].values)
        fig, ax = plt.subplots(
            nrows=len(organs), ncols=len(params),
            figsize=(panel_width * len(params), panel_height * len(organs))
        )
        if len(organs) == 1:
            ax = np.expand_dims(ax, axis=0)
        for j, param in enumerate(params):
            summary_table_param = self.best_model_by_partition(
                partition_select=partition_select,
                metric_select=metric_select,
                cv_mode="mean",
                subset=subset,
                return_run_only=False,
                grouping=["organ", param, param_hue]
            )
            summary_table_param.sort_values([param, param_hue])
            for i, organ in enumerate(organs):
                summary_table = summary_table_param.loc[summary_table_param["organ"].values == organ, :]
                # Plot each metric:
                ycol = f"{partition_show}_{metric_select}"
                if len(organs) == 1 and len(params) == 1:
                    ax = np.array([ax])
                sns.boxplot(
                    x=param, hue=param_hue, y=ycol,
                    data=summary_table, ax=ax[i, j] if len(ax.shape) == 2 else ax[i] if len(ax.shape) == 1 else ax
                )
                if show_swarm:
                    sns.swarmplot(
                        x=param, hue=param_hue, y=ycol,
                        data=summary_table, ax=ax[i, j] if len(ax.shape) == 2 else ax[i] if len(ax.shape) == 1 else ax
                    )
                if j == 0:
                    if len(ax.shape) == 2:
                        ax[i, j].set_ylabel(organ + "\n" + ycol)
                    else:
                        ax[i].set_ylabel(organ + "\n" + ycol)
        return fig, ax

    def plot_training_history(
            self,
            metric_select: str,
            metric_show: str,
            partition_select: str = "val",
            subset: dict = {},
            cv_key: Union[str, None] = None,
            log_loss: bool = False
    ):
        """
        Plot train and validation loss during training and learning rate reduction for each organ

        The partition that is shown in train+val by default because these are the only ones recorded during training.

        :param metric_select: metric to select best model by
        :param metric_show: metric to show as function of training progress, together with loss and learing rate.
        :param partition_select: "train" or "eval" or "test" partition of data to select fit by.
        :param metric_select: Metric to select fit by.
        :param cv_key: Index of cross-validation to plot training history for.
        :param log_loss:
        :return:
        """
        import seaborn as sns
        import matplotlib.pyplot as plt
        panel_width = 5
        panel_height = 3

        organs = np.unique(self.summary_tab["organ"].values)
        fig, ax = plt.subplots(
            nrows=len(organs), ncols=3,
            figsize=(panel_width * 3, panel_height * len(organs))
        )
        if len(organs) == 1:
            ax = np.expand_dims(ax, axis=0)
        for i, organ in enumerate(organs):
            model_gs_id, _, _ = self.get_best_model_ids(
                tab=self.summary_tab,
                partition_select=partition_select,
                metric_select=metric_select,
                cv_mode="mean",
                subset=dict(list(subset.items()) + [("organ", organ)]),
            )
            if cv_key is None:
                sns_data = []
                for run in list(np.unique(
                        self.summary_tab.loc[self.summary_tab["model_gs_id"].values == model_gs_id, "run"].values
                )):
                    sns_data_temp = pandas.DataFrame(self.histories[run])
                    sns_data_temp["epoch"] = np.arange(0, sns_data_temp.shape[0])
                    sns_data_temp["cv"] = run[-1]
                    sns_data.append(sns_data_temp)
                sns_data = pandas.concat(sns_data, axis=0)
            else:
                cv = cv_key
                sns_data = pandas.DataFrame(self.histories[f"{model_gs_id}_{cv}"])
                sns_data["epoch"] = np.arange(0, sns_data.shape[0])
                sns_data["cv"] = cv

            # loss
            sns_data_loss = pandas.concat([pandas.DataFrame({
                "epoch": sns_data["epoch"].values,
                "cv": sns_data["cv"].values,
                "loss": np.log(sns_data[x].values) if log_loss else sns_data[x].values,
                "partition": x
            }) for i, x in enumerate(["loss", "val_loss"])])
            sns.lineplot(
                x="epoch", y="loss", style="partition", hue="cv",
                data=sns_data_loss, ax=ax[i, 0]
            )
            ax[i, 0].set_ylabel(organ + "\nloss")
            ax[i, 0].legend_.remove()

            # metric
            if metric_show not in sns_data.columns:
                raise ValueError(f"metric {metric_show} not found in {sns_data.columns}")
            sns_data_metric = pandas.concat([pandas.DataFrame({
                "epoch": sns_data["epoch"].values,
                "cv": sns_data["cv"].values,
                metric_show: sns_data[metric_show].values,
                "partition": x
            }) for i, x in enumerate([metric_show, f"val_{metric_show}"])])
            sns.lineplot(
                x="epoch", y=metric_show, style="partition", hue="cv",
                data=sns_data_metric, ax=ax[i, 1]
            )
            ax[i, 1].set_ylabel(organ + "\n" + metric_show)
            ax[i, 1].legend_.remove()

            # lr
            sns_data_lr = pandas.DataFrame({
                "epoch": sns_data["epoch"].values,
                "cv": sns_data["cv"].values,
                "lr": np.log(sns_data["lr"].values) / np.log(10)
            })
            sns.lineplot(
                x="epoch", y="lr", hue="cv",
                data=sns_data_lr, ax=ax[i, 2]
            )
            ax[i, 2].set_ylabel("log10 learning rate")
            ax[i, 2].legend_.remove()
        return fig, ax

    def write_best_hyparam(
            self,
            write_path,
            subset: dict = {},
            partition: str = "test",
            metric: str = "custom_negll",
            cvs: Union[None, List[int]] = None
    ):
        best_model_id = self.get_best_model_ids(
            tab=self.summary_tab,
            subset=subset,
            partition_select=partition,
            metric_select=metric,
        )[0]

        if best_model_id is not None:
            if cvs is None:
                file_path_base = os.path.join(
                    self.gs_dirs[best_model_id],
                    self.gs_keys[best_model_id],
                    'results',
                    best_model_id,
                )
            else:
                file_path_base = os.path.join(
                    self.gs_dirs[f"{best_model_id}_cv{cvs[0]}"],
                    f"{best_model_id}_cv{cvs[0]}",
                )

            # Read model hyperparameter
            with open(f"{file_path_base}_model_hyperparam.pickle", 'rb') as file:
                hyparam_model = pickle.load(file)

            # Read optimizer hyperparameter
            with open(f"{file_path_base}_hyperparam.pickle", 'rb') as file:
                hyparam_optim = pickle.load(file)
            # Write both hyperparameter dicts
            with open(os.path.join(write_path, f"{best_model_id[:-12]}_best_hyperparam.txt"), 'w') as file:
                file.write(json.dumps({"model": hyparam_model, "optimizer": hyparam_optim}))
        return


class SummarizeGridsearchCelltype(GridsearchContainer):
    loss_idx: int
    acc_idx: int

    def __init__(
            self,
            source_path: dict,
            cv: bool,
            model_id_len: int = 3
    ):
        super(SummarizeGridsearchCelltype, self).__init__(
            source_path=source_path,
            cv=cv
        )
        self.model_id_len = model_id_len

    def load_ontology_names(
            self,
            run_id: str
    ):
        """
        Loads ontology ids from a specific model of a previously loaded grid search.

        :param run_id:
        :return:
        """
        fn = os.path.join(self.gs_dirs[run_id], f"{run_id}_ontology_names.pickle")
        if not os.path.isfile(fn):
            raise FileNotFoundError(f"file {run_id}_ontology_names.pickle not found")
        with open(fn, 'rb') as f:
            ids = pickle.load(f)
        return ids

    def create_summary_tab(self):
        metrics = list(self.evals.values())[0]['val'].keys()
        self.summary_tab = pandas.DataFrame(dict(
            list({
                "depth":       [id_i.split("_")[self.model_id_len + 0] for id_i in self.run_ids],  # noqa: E241
                "width":       [id_i.split("_")[self.model_id_len + 1] for id_i in self.run_ids],  # noqa: E241
                "lr":          [id_i.split("_")[self.model_id_len + 2] for id_i in self.run_ids],  # noqa: E241
                "dropout":     [id_i.split("_")[self.model_id_len + 3] for id_i in self.run_ids],  # noqa: E241
                "l1":          [id_i.split("_")[self.model_id_len + 4] for id_i in self.run_ids],  # noqa: E241
                "l2":          [id_i.split("_")[self.model_id_len + 5] for id_i in self.run_ids],  # noqa: E241
                "cv":          [id_i.split("_")[-1] if self.cv else "1" for id_i in self.run_ids],  # noqa: E241
                "model":       ["_".join(id_i.split("_")[:self.model_id_len]) for id_i in self.run_ids],  # noqa: E241
                "organism":    [id_i.split("_")[1].split("-")[0] for id_i in self.run_ids],  # noqa: E241
                "organ":       [id_i.split("_")[1].split("-")[1] for id_i in self.run_ids],  # noqa: E241
                "model_type":  [  # noqa: E241
                    "linear" if (id_i.split("_")[1].split("-")[2] == "mlp" and
                                 id_i.split("_")[1].split("-")[3].split(".")[1] == "0")
                    else id_i.split("_")[1].split("-")[2]
                    for id_i in self.run_ids
                ],
                "version":     [id_i.split("_")[1].split("-")[3] for id_i in self.run_ids],  # noqa: E241
                "model_gs_id": ["_".join(id_i.split("_")[:(self.model_id_len + 6)]) for id_i in self.run_ids],  # noqa: E241
                "run":         self.run_ids,  # noqa: E241
            }.items()) +
            list(dict([("train_" + m, [self.evals[x]["train"][m] for x in self.run_ids]) for m in metrics]).items()) +
            list(dict([("test_" + m, [self.evals[x]["test"][m] for x in self.run_ids]) for m in metrics]).items()) +
            list(dict([("val_" + m, [self.evals[x]["val"][m] for x in self.run_ids]) for m in metrics]).items()) +
            list(dict([("all_" + m, [self.evals[x]["all"][m] for x in self.run_ids]) for m in metrics]).items())
        ))
        if self.summary_tab.shape[0] == 0:
            raise ValueError("summary_tab was empty")

    def best_model_celltype(
            self,
            subset: dict = {},
            partition: str = "val",
            metric: str = "loss",
            cvs: Union[None, List[int]] = None
    ):
        model_id, _, _ = self.get_best_model_ids(
            tab=self.summary_tab,
            partition_select=partition,
            metric_select=metric,
            cv_mode="mean",
            subset=subset,
        )
        if model_id is not None:
            if cvs is not None:
                fns = [
                    os.path.join(self.gs_dirs[f"{model_id}_cv{x}"], self.gs_keys[f"{model_id}_cv{x}"], "results", f"{model_id}_cv{x}")
                    for x in cvs
                ]
            else:
                fns = [os.path.join(self.gs_dirs[model_id], self.gs_keys[model_id], "results", model_id)]
            covar = [pandas.read_csv(f"{x}_covar.csv") for x in fns]
            return model_id, covar
        else:
            return None, [None]

    def plot_best(
            self,
            rename_levels=[],
            partition_select: str = "val",
            partition_show: str = "test",
            metric_select: str = "acc",
            metric_show: str = "acc",
            collapse_cv: str = "max",
            vmin=None,
            vmax=None,
            height_fig=7,
            width_fig=7
    ):
        """
        Plot accuracy or other metric heatmap by organ and model type.

        :param rename_levels:
        :param partition_select:
        :param partition_show:
        :param metric_select:
        :param metric_show:
        :param collapse_cv:
        :param vmin:
        :param vmax:
        :param height_fig:
        :param width_fig:
        :return:
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        if self.summary_tab is None:
            self.create_summary_tab()

        # Choose the best over categories based on mean loss in CV.
        # Keep variation across CV.
        sns_tab = self.best_model_by_partition(
            partition_select=partition_select,
            metric_select=metric_select,
            return_run_only=False,
            grouping=["organ", "model_type"]
        )
        for rename_level in rename_levels:
            levels_new = sns_tab[rename_level[0]].values
            levels_new[levels_new == rename_level[1]] = rename_level[2]
            sns_tab[rename_level[0]] = levels_new

        # Build figure.
        organs = np.unique(sns_tab["organ"].values)
        model_types = np.unique(sns_tab["model_type"].values)
        hm = np.zeros((len(organs), len(model_types))) + np.nan
        mask = np.isnan(hm)
        for i, m in enumerate(model_types):
            for j, o in enumerate(organs):
                data_temp = sns_tab.loc[
                    np.logical_and(
                        sns_tab["model_type"].values == m,
                        sns_tab["organ"].values == o
                    ), f"{partition_show}_{metric_show}"
                ]
                if data_temp.shape[0] > 0:
                    if self.cv:
                        if collapse_cv == "mean":
                            hm[j, i] = np.mean(data_temp.values)
                        elif collapse_cv == "median":
                            hm[j, i] = np.median(data_temp.values)
                        elif collapse_cv == "max":
                            hm[j, i] = np.max(data_temp.values)
                        elif collapse_cv == "min":
                            hm[j, i] = np.min(data_temp.values)
                        else:
                            raise ValueError(f"collapse_cv {collapse_cv} not recognized")
                        mask[j, i] = False
                    else:
                        hm[j, i] = data_temp.values[0]
                        mask[j, i] = False
        if vmin is not None:
            hm = np.maximum(hm, np.asarray(vmin))
        if vmax is not None:
            hm = np.minimum(hm, np.asarray(vmin))
        sns_data_heatmap = pandas.DataFrame(
            hm, index=organs, columns=model_types
        )
        fig, axs = plt.subplots(1, 1, figsize=(height_fig, width_fig))
        with sns.axes_style("dark"):
            axs = sns.heatmap(
                sns_data_heatmap,  # mask=mask,
                annot=True, fmt=".2f",
                ax=axs, vmin=0, vmax=1,
                xticklabels=True, yticklabels=True,
                cbar_kws={'label': f"{partition_show}_{metric_show}"},
                cmap=None
            )
        return fig, axs, sns_data_heatmap

    def plot_best_classwise_heatmap(
            self,
            organ: str,
            organism: str,
            datapath: str,
            store_format: str,
            targetpath: str,
            configpath: str,
            partition_select: str = "val",
            metric_select: str = "custom_cce_agg",
            metric_show: str = "f1",
            collapse_cv: str = "mean",
            min_cells: int = 10,
            height_fig: int = 7,
            width_fig: int = 7
    ):
        """
        Plot evaluation metric heatmap for specified organ by cell classes and model types.

        :param organ: Organ to plot in heatmap.
        :param organism: Species that the gridsearch was run on
        :param datapath: Path to the local sfaira data repository
        :param store_format:
        :param targetpath:
        :param configpath:
        :param partition_select: Based on which partition to select the best model
            - train
            - val
            - test
            - all
        :param metric_select: Based on which metric to select the best model
            - loss
            - accuracy
            - custom_cce_agg
            - acc_agg
            - f1
            - tpr
            - fpr
        :param metric_show: Which classwise metric to plot.
            - accuracy
            - f1
        :param collapse_cv: How to collapse over the single cv runs.
        :param min_cells: Minimum number of cells of a type must be present in the whole dataset for that class to be included in the plot.
        :param height_fig: Figure height.
        :param width_fig: Figure width.
        :return: fig, axs, sns_data_heatmap
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        def f1(yhat, ytrue):
            """
            Class wise F1.

            :param yhat:
            :param ytrue:
            :return:
            """
            def _tp(yhat, ytrue):
                """
                Class wise true positive count.

                :param yhat:
                :param ytrue:
                :return:
                """
                yhat_true = np.asarray(yhat == np.max(yhat, axis=1, keepdims=True), dtype="float32")
                return np.sum(yhat_true * ytrue, axis=0)

            def _fp(yhat, ytrue):
                """
                Class wise false positive count.

                :param yhat:
                :param ytrue:
                :return:
                """
                yhat_true = np.asarray(yhat == np.max(yhat, axis=1, keepdims=True), dtype="float32")
                return np.sum(yhat_true * (1. - ytrue), axis=0)

            def _fn(yhat, ytrue):
                """
                Class wise false negative count.

                :param yhat:
                :param ytrue:
                :return:
                """
                yhat_true = np.asarray(yhat < np.max(yhat, axis=1, keepdims=True), dtype="float32")
                return np.sum(yhat_true * ytrue, axis=0)
            precision = _tp(yhat, ytrue) / (_tp(yhat, ytrue) + _fp(yhat, ytrue))
            recall = _tp(yhat, ytrue) / (_tp(yhat, ytrue) + _fn(yhat, ytrue))
            return 2 * 1 / (1 / precision + 1 / recall)

        if self.summary_tab is None:
            self.create_summary_tab()

        # Choose the best over categories based on mean loss in CV.
        # Keep variation across CV.
        sns_tab = self.best_model_by_partition(
            partition_select=partition_select,
            metric_select=metric_select,
            return_run_only=False,
            grouping=["organ", "model_type"]
        )
        sns_tab = sns_tab[sns_tab['organ'] == organ]

        store = load_store(cache_path=datapath, store_format=store_format)
        store.load_config(configpath)
        store.subset(attr_key="id", values=[k for k in store.indices.keys()
                                            if 'cell_ontology_class' in store.adata_by_key[k].obs.columns])
        store.subset(attr_key="cellontology_class", excluded_values=[
            store._adata_ids_sfaira.unknown_metadata_identifier,
            store._adata_ids_sfaira.not_a_cell_celltype_identifier,
        ])
        cu = CelltypeUniverse(
            cl=OntologyCl(branch="v2021-02-01"),
            uberon=OCS.organ,
        )
        cu.load_target_universe(targetpath)
        cell_counts = store.obs['cell_ontology_class'].value_counts().to_dict()
        celltypelist = list(cell_counts.keys()).copy()
        leaves = cu.onto_cl.convert_to_name(cu.onto_cl.leaves)
        for k in celltypelist:
            leafnodes = cu.onto_cl.convert_to_name(cu.onto_cl.map_to_leaves(node=k, return_type="ids", include_self=True))
            # Replace count on intermediate nodes with counts over leaves
            if k not in leaves:
                for leaf in leaves:
                    if leaf not in cell_counts.keys():
                        cell_counts[leaf] = 0
                    cell_counts[leaf] += 1. / len(leafnodes)
                del cell_counts[k]

        # Compute class-wise metrics
        vals = []
        for i, run_id in enumerate(sns_tab["run"].values):
            yhat = self.load_y(hat_or_true='hat', run_id=run_id)
            ytrue = self.load_y(hat_or_true='true', run_id=run_id)
            if metric_show == "acc":
                m = accuracy(yhat, ytrue)
            elif metric_show == "f1":
                m = f1(yhat, ytrue)
            else:
                raise ValueError(f"did not recognize metric_show {metric_show}")
            vals.append(m)
        sns_tab[f"{metric_show}_classwise"] = vals

        # Build figure.
        model_types = sns_tab["model_type"].unique()
        model_types.sort()
        classes = self.load_ontology_names(run_id=sns_tab["run"].values[0])
        hm = np.zeros((len(classes), len(model_types))) + np.nan
        # mask = np.isnan(hm)
        for i, m in enumerate(model_types):
            data_temp = np.vstack(sns_tab.loc[sns_tab["model_type"].values == m, f"{metric_show}_classwise"].values)
            if data_temp.shape[0] > 0:
                if self.cv:
                    if collapse_cv == "mean":
                        hm[:, i] = np.nanmean(data_temp, axis=0)
                    elif collapse_cv == "median":
                        hm[:, i] = np.nanmedian(data_temp, axis=0)
                    elif collapse_cv == "max":
                        hm[:, i] = np.nanmax(data_temp, axis=0)
                    elif collapse_cv == "min":
                        hm[:, i] = np.nanmin(data_temp, axis=0)
                    else:
                        raise ValueError(f"collapse_cv {collapse_cv} not recognized")
                else:
                    hm[:, i] = data_temp.values[0]
        n_cells = []
        for c in classes:
            if c in cell_counts.keys():
                n_cells.append(np.round(cell_counts[c]))
            else:
                warnings.warn(f"Celltype {c} from cell ontology not found in {organism} {organ} dataset")
                n_cells.append(np.nan)
        n_cells = np.array(n_cells)[:, None]
        sns_data_heatmap = pandas.DataFrame(
            np.hstack((n_cells, hm)),
            index=classes,
            columns=['Number of cells in whole dataset'] + list(model_types)
        )
        sns_data_heatmap = sns_data_heatmap[sns_data_heatmap['Number of cells in whole dataset'] >= min_cells]
        mask = np.zeros(sns_data_heatmap.shape).astype(bool)
        mask[:, 0] = True
        with sns.axes_style("dark"):
            fig, axs = plt.subplots(1, 1, figsize=(width_fig, height_fig))
            axs = sns.heatmap(
                sns_data_heatmap, mask=mask,
                annot=True, fmt=".2f",
                ax=axs, vmin=0, vmax=1,
                xticklabels=True, yticklabels=True,
                cbar_kws={'label': f"test_{metric_show}"},
                cmap=None
            )
            axs = sns.heatmap(
                data=sns_data_heatmap, mask=~mask,
                annot=True, fmt=".0f",
                ax=axs, alpha=0,
                xticklabels=True, yticklabels=True,
                annot_kws={"color": "black"},
                cbar=False
            )
        return fig, axs, sns_data_heatmap

    def plot_best_classwise_scatter(
            self,
            organ: str,
            organism: str,
            datapath: str,
            store_format: str,
            targetpath: str,
            configpath: str,
            partition_select: str = "val",
            metric_select: str = "custom_cce_agg",
            metric_show: str = "f1",
            collapse_cv: str = "mean",
            min_cells: int = 10,
            height_fig: int = 7,
            width_fig: int = 7,
            annotate_thres_ncells: int = 1000,
            annotate_thres_f1: float = 0.5,
    ):
        """
        Plot evaluation metric scatterplot for specified organ by cell classes and model types.

        :param organ: Organ to plot in heatmap.
        :param organism: Organism that the gridsearch was run on
        :param datapath: Path to the local sfaira data repository
        :param store_format:
        :param targetpath:
        :param configpath:
        :param partition_select: Based on which partition to select the best model
            - train
            - val
            - test
            - all
        :param metric_select: Based on which metric to select the best model
            - loss
            - accuracy
            - custom_cce_agg
            - acc_agg
            - f1
            - tpr
            - fpr
        :param metric_show: Which classwise metric to plot.
            - accuracy
            - f1
        :param collapse_cv: How to collapse over the single cv runs.
        :param min_cells: Minimum number of cells of a type must be present in the whole dataset for that class to be included in the plot.
        :param height_fig: Figure height.
        :param width_fig: Figure width.
        :param annotate_thres_ncells:
        :param annotate_thres_f1:
        :return: fig, axs, sns_data_scatter
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        if self.summary_tab is None:
            self.create_summary_tab()

        # Choose the best over categories based on mean loss in CV.
        # Keep variation across CV.
        sns_tab = self.best_model_by_partition(
            partition_select=partition_select,
            metric_select=metric_select,
            return_run_only=False,
            grouping=["organ", "model_type"]
        )
        sns_tab = sns_tab[sns_tab['organ'] == organ]

        store = load_store(cache_path=datapath, store_format=store_format)
        store.load_config(configpath)
        store.subset(attr_key="id", values=[k for k in store.indices.keys()
                                            if 'cell_ontology_id' in store.adata_by_key[k].obs.columns])
        store.subset(attr_key="cellontology_class", excluded_values=[
            store._adata_ids_sfaira.unknown_metadata_identifier,
            store._adata_ids_sfaira.not_a_cell_celltype_identifier,
        ])
        cu = CelltypeUniverse(
            cl=OntologyCl(branch="v2021-02-01"),
            uberon=OCS.organ,
        )
        cu.load_target_universe(targetpath)
        cell_counts = store.obs['cell_ontology_class'].value_counts().to_dict()
        celltypelist = list(cell_counts.keys()).copy()
        leaves = cu.onto_cl.convert_to_name(cu.onto_cl.leaves)
        for k in celltypelist:
            leafnodes = cu.onto_cl.convert_to_name(cu.onto_cl.map_to_leaves(node=k, return_type="ids", include_self=True))
            # Replace count on intermediate nodes with counts over leaves
            if k not in leaves:
                for leaf in leaves:
                    if leaf not in cell_counts.keys():
                        cell_counts[leaf] = 0
                    cell_counts[leaf] += 1. / len(leafnodes)
                del cell_counts[k]

        # Compute class-wise metrics
        vals = []
        for i, run_id in enumerate(sns_tab["run"].values):
            yhat = self.load_y(hat_or_true='hat', run_id=run_id)
            ytrue = self.load_y(hat_or_true='true', run_id=run_id)
            if metric_show == "acc":
                m = accuracy(yhat, ytrue)
            elif metric_show == "f1":
                m = f1(yhat, ytrue)
            else:
                raise ValueError("did not recognize metric_show %s" % metric_show)
            vals.append(m)
        sns_tab[f"{metric_show}_classwise"] = vals

        # Build figure.
        model_types = sns_tab["model_type"].unique()
        classes = self.load_ontology_names(run_id=sns_tab["run"].values[0])
        hm = np.zeros((len(classes), len(model_types))) + np.nan
        # mask = np.isnan(hm)
        for i, m in enumerate(model_types):
            data_temp = np.vstack(sns_tab.loc[sns_tab["model_type"].values == m, f"{metric_show}_classwise"].values)
            if data_temp.shape[0] > 0:
                if self.cv:
                    if collapse_cv == "mean":
                        hm[:, i] = np.nanmean(data_temp, axis=0)
                    elif collapse_cv == "median":
                        hm[:, i] = np.nanmedian(data_temp, axis=0)
                    elif collapse_cv == "max":
                        hm[:, i] = np.nanmax(data_temp, axis=0)
                    elif collapse_cv == "min":
                        hm[:, i] = np.nanmin(data_temp, axis=0)
                    else:
                        raise ValueError(f"collapse_cv {collapse_cv} not recognized")
                else:
                    hm[:, i] = data_temp.values[0]
        n_cells = []
        for c in classes:
            if c in cell_counts.keys():
                n_cells.append(np.round(cell_counts[c]))
            else:
                warnings.warn(f"Celltype {c} from cell cu not found in {organism} {organ} dataset")
                n_cells.append(np.nan)
        n_cells = np.array(n_cells)[:, None]
        sns_data_scatter = pandas.DataFrame(
            np.hstack((n_cells, hm)),
            index=classes,
            columns=['Number of cells in whole dataset'] + list(model_types)
        )
        sns_data_scatter = sns_data_scatter[sns_data_scatter['Number of cells in whole dataset'] >= min_cells]
        sns_data_scatter = pandas.melt(sns_data_scatter,
                                       id_vars=['Number of cells in whole dataset'],
                                       value_vars=list(model_types),
                                       var_name='Model type',
                                       value_name='Classwise f1 score',
                                       ignore_index=False
                                       )

        with sns.axes_style("dark"):
            fig, axs = plt.subplots(1, 1, figsize=(width_fig, height_fig))
            axs = sns.scatterplot(x='Number of cells in whole dataset',
                                  y='Classwise f1 score',
                                  style='Model type',
                                  alpha=0.8,
                                  data=sns_data_scatter,
                                  ax=axs
                                  )
            for line in range(0, sns_data_scatter.shape[0]):
                if (sns_data_scatter['Number of cells in whole dataset'][line] > annotate_thres_ncells) \
                        and (sns_data_scatter['Classwise f1 score'][line] > annotate_thres_f1):
                    axs.text(sns_data_scatter['Number of cells in whole dataset'][line] + 100,
                             sns_data_scatter['Classwise f1 score'][line],
                             sns_data_scatter.index[line],
                             horizontalalignment='left',
                             size='medium',
                             color='black',
                             weight='semibold'
                             )

        return fig, axs, sns_data_scatter


class SummarizeGridsearchEmbedding(GridsearchContainer):
    loss_idx: int
    mse_idx: int

    def __init__(
            self,
            source_path: dict,
            cv: bool,
            loss_idx: int = 0,
            mse_idx: int = 1,
            model_id_len: int = 3
    ):
        super(SummarizeGridsearchEmbedding, self).__init__(
            source_path=source_path,
            cv=cv
        )
        self.loss_idx = loss_idx
        self.mse_idx = mse_idx
        self.model_id_len = model_id_len

    def create_summary_tab(self):
        metrics = list(self.evals.values())[0]['val'].keys()
        self.summary_tab = pandas.DataFrame(dict(
            list({
                "depth":         [id_i.split("_")[self.model_id_len + 0] for id_i in self.run_ids],  # noqa: E241
                "width":         [id_i.split("_")[self.model_id_len + 1] for id_i in self.run_ids],  # noqa: E241
                "lr":            [id_i.split("_")[self.model_id_len + 2] for id_i in self.run_ids],  # noqa: E241
                "dropout":       [id_i.split("_")[self.model_id_len + 3] for id_i in self.run_ids],  # noqa: E241
                "l1":            [id_i.split("_")[self.model_id_len + 4] for id_i in self.run_ids],  # noqa: E241
                "l2":            [id_i.split("_")[self.model_id_len + 5] for id_i in self.run_ids],  # noqa: E241
                "cv":            [id_i.split("_")[-1] if self.cv else "1" for id_i in self.run_ids],  # noqa: E241
                "model":         ["_".join(id_i.split("_")[:self.model_id_len]) for id_i in self.run_ids],  # noqa: E241
                "organism":      [id_i.split("_")[1].split("-")[0] for id_i in self.run_ids],  # noqa: E241
                "organ":         [id_i.split("_")[1].split("-")[1] for id_i in self.run_ids],  # noqa: E241
                "model_type":    [id_i.split("_")[1].split("-")[2] for id_i in self.run_ids],  # noqa: E241
                "version":       [id_i.split("_")[1].split("-")[3] for id_i in self.run_ids],  # noqa: E241
                "model_gs_id":   ["_".join(id_i.split("_")[:(self.model_id_len + 6)]) for id_i in self.run_ids],  # noqa: E241
                "run":           self.run_ids,  # noqa: E241
            }.items()) +
            list(dict([("train_" + m, [self.evals[x]["train"][m] if m in self.evals[x]["train"].keys() else
                                       self.evals[x]["train"]['neg_ll_' + m] for x in self.run_ids])
                       for m in metrics]).items()) +
            list(dict([("test_" + m, [self.evals[x]["test"][m] if m in self.evals[x]["test"].keys() else
                                      self.evals[x]["test"]['neg_ll_' + m] for x in self.run_ids])
                       for m in metrics]).items()) +
            list(dict([("val_" + m, [self.evals[x]["val"][m] if m in self.evals[x]["val"].keys()
                                     else self.evals[x]["val"]['neg_ll_' + m] for x in self.run_ids])
                       for m in metrics]).items()) +
            list(dict([("all_" + m, [self.evals[x]["all"][m] if m in self.evals[x]["all"].keys()
                                     else self.evals[x]["all"]['neg_ll_' + m] for x in self.run_ids])
                       for m in metrics]).items())
        ))
        # TODO: Hacky solution to make sure metrics are called the same in VAE and other models
        rename_dict = {
            "train_neg_ll_custom_mse": "train_custom_mse",
            "train_neg_ll_custom_negll": "train_custom_negll",
            "test_neg_ll_custom_mse": "test_custom_mse",
            "test_neg_ll_custom_negll": "test_custom_negll",
            "val_neg_ll_custom_mse": "val_custom_mse",
            "val_neg_ll_custom_negll": "val_custom_negll",
            "all_neg_ll_custom_mse": "all_custom_mse",
            "all_neg_ll_custom_negll": "all_custom_negll",
        }
        self.summary_tab = self.summary_tab.rename(columns=rename_dict)

        if self.summary_tab.shape[0] == 0:
            raise ValueError("summary_tab was empty")

    def best_model_embedding(
            self,
            subset: dict = {},
            partition: str = "val",
            metric: str = "loss",
            cvs: Union[None, List[int]] = None
    ):
        model_id, _, _ = self.get_best_model_ids(
            tab=self.summary_tab,
            partition_select=partition,
            metric_select=metric,
            cv_mode="mean",
            subset=subset,
        )
        if model_id is not None:
            if cvs is not None:
                fns = [
                    os.path.join(self.gs_dirs[f"{model_id}_cv{x}"], f"{model_id}_cv{x}")
                    for x in cvs
                ]
            else:
                fns = [os.path.join(self.gs_dirs[model_id], self.gs_keys[model_id], "results", model_id)]
            embedding = [np.load(f"{x}_embedding.npy") for x in fns]
            covar = [pandas.read_csv(f"{x}_covar.csv") for x in fns]
            return model_id, embedding, covar
        else:
            return None, [None], [None]

    def plot_best(
            self,
            rename_levels=[],
            partition_select: str = "val",
            partition_show: str = "test",
            metric_select: str = "ll",
            metric_show: str = "ll",
            collapse_cv: str = "min",
            vmin=None,
            vmax=None,
            height_fig=7,
            width_fig=7
    ):
        """

        :param rename_levels:
        :param partition_select:
        :param partition_show:
        :param metric_select:
        :param metric_show:
        :param collapse_cv:
        :param vmin:
        :param vmax:
        :param height_fig:
        :param width_fig:
        :return:
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        if self.summary_tab is None:
            self.create_summary_tab()

        # Choose the best over categories based on mean loss in CV.
        # Keep variation across CV.
        sns_tab = self.best_model_by_partition(
            partition_select=partition_select,
            metric_select=metric_select,
            return_run_only=False,
            grouping=["organ", "model_type"]
        )
        for rename_level in rename_levels:
            levels_new = sns_tab[rename_level[0]].values
            levels_new[levels_new == rename_level[1]] = rename_level[2]
            sns_tab[rename_level[0]] = levels_new

        # Build figure.
        organs = np.unique(sns_tab["organ"].values)
        model_types = np.unique(sns_tab["model_type"].values)
        hm = np.zeros((len(organs), len(model_types))) + np.nan
        mask = np.isnan(hm)
        for i, m in enumerate(model_types):
            for j, o in enumerate(organs):
                data_temp = sns_tab.loc[
                    np.logical_and(
                        sns_tab["model_type"].values == m,
                        sns_tab["organ"].values == o
                    ), f"{partition_show}_{metric_show}"
                ]
                if data_temp.shape[0] > 0:
                    if self.cv:
                        if collapse_cv == "mean":
                            hm[j, i] = np.mean(data_temp.values)
                        elif collapse_cv == "median":
                            hm[j, i] = np.median(data_temp.values)
                        elif collapse_cv == "max":
                            hm[j, i] = np.max(data_temp.values)
                        elif collapse_cv == "min":
                            hm[j, i] = np.min(data_temp.values)
                        else:
                            raise ValueError("collapse_cv % s not recognized" % collapse_cv)
                    else:
                        hm[j, i] = data_temp.values[0]
                        mask[j, i] = False
        if vmin is not None:
            hm = np.maximum(hm, np.asarray(vmin))
        if vmax is not None:
            hm = np.minimum(hm, np.asarray(vmin))
        sns_data_heatmap = pandas.DataFrame(
            hm, index=organs, columns=model_types
        )
        fig, axs = plt.subplots(1, 1, figsize=(height_fig, width_fig))
        with sns.axes_style("dark"):
            axs = sns.heatmap(
                sns_data_heatmap,
                annot=True, fmt=".2f",
                ax=axs,
                xticklabels=True, yticklabels=True,
                cbar_kws={'label': f"{partition_show}_{metric_show}"}
            )
        return fig, axs, sns_data_heatmap

    from typing import Union, List

    def get_gradients_by_celltype(
            self,
            model_organ: str,
            data_organ: str,
            organism: Union[str, None],
            genome: Union[str, None, dict],
            model_type: Union[str, List[str]],
            metric_select: str,
            data_source: str,
            datapath,
            gene_type: str = "protein_coding",
            configpath: Union[None, str] = None,
            store_format: Union[None, str] = None,
            test_data=True,
            partition_select: str = "val",
            ignore_cache=False,
            min_cells=10,
    ):
        """
        Compute gradients across latent units with respect to input features for each cell type.

        :param model_organ:
        :param data_organ:
        :param organism:
        :param model_type:
        :param metric_select:
        :param datapath:
        :param test_data:
        :param partition_select:
        :param ignore_cache:
        :param min_cells:
        :return: (cell types, input features) cumulative gradients
        """
        model_id, run_id, _ = self.get_best_model_ids(
            tab=self.summary_tab,
            metric_select=metric_select,
            partition_select=partition_select,
            subset={
                "organ": model_organ,
                "model_type": model_type,
            }
        )
        resultspath = self.gs_dirs[run_id]
        if os.path.isfile(os.path.join(resultspath, f'{model_id}_grads.pickle')) and not ignore_cache:
            print('Load gradients from cached file...')
            with open(os.path.join(resultspath, f'{model_id}_grads.pickle'), 'rb') as f:
                gradients_raw = pickle.load(f)
        else:
            print('Compute gradients (1/3): load data')
            # load data
            if data_source == "store":
                if genome is not None:
                    warnings.warn("Using data_source='store', the provided genome will be ignored")
                store = load_store(cache_path=datapath, store_format=store_format)
                store.load_config(configpath)
                store.subset(attr_key="id", values=[k for k in store.indices.keys()
                                                    if 'cell_ontology_id' in store.adata_by_key[k].obs.columns])
                store.subset(attr_key="cellontology_class", excluded_values=[
                    store._adata_ids_sfaira.unknown_metadata_identifier,
                    store._adata_ids_sfaira.not_a_cell_celltype_identifier,
                ])
                adatas = store.adata_sliced
                # Load into memory:
                for k in adatas.keys():
                    adatas[k] = adatas[k].to_memory()
                adata = adatas[list(adatas.keys())[0]]
                if len(adatas.keys()) > 0:
                    adata = adata.concatenate(*[adatas[k] for k in list(adatas.keys())[1:]])
            elif data_source == "universe":
                if configpath is not None or store_format is not None:
                    warnings.warn("Using data_source='universe', the provided configpath and store_format will be ignored")
                u = Universe(data_path=datapath)
                if organism is not None:
                    u.subset("organism", organism)
                if data_organ is not None:
                    u.subset("organ", data_organ)
                u.load(allow_caching=False)
                u.streamline_features(match_to_release=genome, subset_genes_to_type=gene_type)
                u.streamline_metadata()
                adata = u.adata
            else:
                raise ValueError("data_source has to be 'universe' or 'store'")

            print('Compute gradients (2/3): load embedding')
            zoo = ModelZoo()
            zoo.model_id = "_".join(model_id.split("_")[:3])
            embedding = EstimatorKerasEmbedding(
                data=adata,
                model_dir="",
                model_id=model_id,
                model_topology=zoo.topology_container
            )
            embedding.init_model()
            embedding.model.training_model.load_weights(os.path.join(resultspath, f'{model_id}_weights.h5'))

            # compute gradients
            print('Compute gradients (3/3): cumulate gradients')
            gradients_raw = embedding.compute_gradients_input(test_data=test_data, batch_size=256, per_celltype=True)
            with open(os.path.join(resultspath, f'{model_id}_grads.pickle'), 'wb') as f:
                pickle.dump(gradients_raw, f, pickle.HIGHEST_PROTOCOL)
            print('Gradients saved to cache file!')

        # filter by minimum number cells min_cells
        filtered_grads = {}
        celltypes = []
        for celltype in gradients_raw['gradients'].keys():
            if gradients_raw['counts'][celltype] > min_cells:
                filtered_grads.update({celltype: gradients_raw['gradients'][celltype]})
                celltypes.append(celltype)

        return np.concatenate([
            np.mean(a, axis=0, keepdims=True)
            for a in list(filtered_grads.values())
        ], axis=0), celltypes

    def plot_gradient_distr(
            self,
            model_organ: str,
            data_organ: str,
            model_type: Union[str, List[str]],
            metric_select: str,
            datapath: str,
            data_source: str,
            organism: Union[str, None] = None,
            genome: Union[str, None] = None,
            configpath: Union[None, str] = None,
            store_format: Union[None, str] = None,
            test_data=True,
            gene_type: str = "protein_coding",
            partition_select: str = "val",
            normalize=True,
            remove_inactive=True,
            min_cells=10,
            bw=0.02,
            xlim=None,
            by_type=True,
            height_fig=7,
            width_fig=7,
            hist=False,
            ignore_cache=False,
            save=None,
    ):
        import seaborn as sns
        import matplotlib.pyplot as plt

        if by_type and isinstance(model_type, list):
            raise ValueError("cannot plot by type and by model")
        if isinstance(model_type, str):
            model_type = [model_type]

        if self.summary_tab is None:
            self.create_summary_tab()

        # best model for each organ and model_type
        avg_grads = {}
        celltypes = {}
        for modelt in model_type:
            avg_grads[modelt], celltypes[modelt] = self.get_gradients_by_celltype(
                model_organ=model_organ,
                data_organ=data_organ,
                organism=organism,
                model_type=modelt,
                metric_select=metric_select,
                genome=genome,
                gene_type=gene_type,
                data_source=data_source,
                datapath=datapath,
                configpath=configpath,
                store_format=store_format,
                test_data=test_data,
                partition_select=partition_select,
                ignore_cache=ignore_cache,
                min_cells=min_cells,
            )

            if normalize:
                avg_grads[modelt] = np.abs(avg_grads[modelt])
                avg_grads[modelt] = (avg_grads[modelt] - np.min(avg_grads[modelt], axis=1, keepdims=True)) /\
                    np.maximum(np.max(avg_grads[modelt], axis=1, keepdims=True) -
                               np.min(avg_grads[modelt], axis=1, keepdims=True), 1e-8)

        fig, axs = plt.subplots(1, 1, figsize=(width_fig, height_fig))

        if len(avg_grads.values()) == 1:
            threshold = np.mean(list(avg_grads.values())[0]) * 0.05
            avg_grads_mask = np.mean(list(avg_grads.values())[0], axis=0) > threshold
            active_grads = list(avg_grads.values())[0][:, avg_grads_mask]
            plt.axvline(threshold, color='k', linestyle='dashed', linewidth=1,
                        label="active gene threshold"
                        )
            plt.axvline(np.mean(active_grads), color='k', linestyle='solid', linewidth=1,
                        label="average gradient\nof active genes")
            print('number of active inputs: ', active_grads.shape[1])

        for k, v in avg_grads.items():
            if by_type:
                v_mask = np.mean(v, axis=0) > threshold
                for i, x in enumerate(v):
                    if remove_inactive:
                        x = x[v_mask]
                    if not hist:
                        sns.kdeplot(x, bw_method=bw, ax=axs)
            else:
                if remove_inactive:
                    threshold = np.mean(v) * 0.05
                    v_mask = np.mean(v, axis=0) > threshold
                    v = v[:, v_mask]
                if not hist:
                    sns.kdeplot(np.asarray(v).flatten(), bw_method=bw, label=k, ax=axs)

        if xlim is not None:
            axs.set_xlim(xlim)
        plt.legend(loc="best")
        plt.xlabel(r'$\rm{mean}_{i=1,...,D} \frac{\partial z_i}{\partial x}$')
        if hist:
            plt.ylabel('# genes')
        plt.tight_layout()
        if save is not None:
            plt.savefig(save)
        plt.show()

    def plot_gradient_cor(
            self,
            model_organ: str,
            data_organ: str,
            model_type: Union[str, List[str]],
            metric_select: str,
            datapath: str,
            data_source: str,
            organism: Union[str, None] = None,
            genome: Union[str, None] = None,
            configpath: Union[None, str] = None,
            store_format: Union[None, str] = None,
            test_data=True,
            gene_type: str = "protein_coding",
            partition_select: str = "val",
            height_fig=7,
            width_fig=7,
            ignore_cache=False,
            min_cells=10,
            by_type=True,
            vmin=0.,
            vmax=1.,
            save=None,
    ):
        """
        Plot correlation heatmap of gradient vectors accumulated on input features between cell types or models.

        :param model_organ:
        :param data_organ:
        :param model_type:
        :param metric_select:
        :param datapath:
        :param configpath:
        :param store_format:
        :param test_data:
        :param partition_select:
        :param height_fig:
        :param width_fig:
        :param ignore_cache:
        :param min_cells:
        :param by_type:
        :param vmin:
        :param vmax:
        :param save:
        :return:
        """
        import seaborn as sns
        import matplotlib.pyplot as plt

        if by_type and isinstance(model_type, list):
            raise ValueError("cannot plot by type and by model")
        if isinstance(model_type, str):
            model_type = [model_type]

        if self.summary_tab is None:
            self.create_summary_tab()

        # best model for each organ and model_type
        avg_grads = {}
        celltypes = {}
        for modelt in model_type:
            avg_grads[modelt], celltypes[modelt] = self.get_gradients_by_celltype(
                model_organ=model_organ,
                data_organ=data_organ,
                organism=organism,
                model_type=modelt,
                metric_select=metric_select,
                genome=genome,
                gene_type=gene_type,
                data_source=data_source,
                datapath=datapath,
                configpath=configpath,
                store_format=store_format,
                test_data=test_data,
                partition_select=partition_select,
                ignore_cache=ignore_cache,
                min_cells=min_cells,
            )

        fig, axs = plt.subplots(1, 1, figsize=(width_fig, height_fig))
        if by_type:
            v = avg_grads[model_type[0]]
            celltypes_coord = celltypes[model_type[0]]
            cormat = pandas.DataFrame(
                np.corrcoef(v),
                index=celltypes_coord,
                columns=celltypes_coord
            )
            sns.heatmap(cormat, vmin=vmin, vmax=vmax, ax=axs)
        else:
            pass

        plt.tight_layout()
        if save is not None:
            plt.savefig(save)
        plt.show()

    def plot_npc(
            self,
            organ,
            topology_version,
            cvs=None
    ):
        """
        Plots the explained variance ration that accumulates explained variation of the latent space’s ordered
        principal components.
        If an embedding file is found that contains z, z_mean, z_var (eg. output from predict_variational() function)
        the model will use z, and not z_mean.
        """
        import matplotlib.pyplot as plt
        if self.summary_tab is None:
            self.create_summary_tab()
        models = np.unique(self.summary_tab["model_type"]).tolist()
        self.summary_tab["topology"] = [x.split("_")[5] for x in self.summary_tab["model_gs_id"].values]

        with plt.style.context("seaborn-whitegrid"):
            plt.figure(figsize=(12, 6))
            for model in models:
                model_id, embedding, covar = self.best_model_embedding(
                    subset={"model_type": model, "organ": organ, "topology": topology_version},
                    partition="val",
                    metric="loss",
                    cvs=cvs,
                )
                if len(embedding[0].shape) == 3:
                    z = embedding[0][0]  # in case of three-dimensional VAE embedding (z, z_mean, z_var), use z
                else:
                    z = embedding[0]
                cov = np.cov(z.T)
                eig_vals, eig_vecs = np.linalg.eig(cov)
                eig_sum = sum(eig_vals)
                var_exp = [(i / eig_sum) for i in sorted(eig_vals, reverse=True)]
                cum_var_exp = np.cumsum([0] + var_exp)
                plt.step(range(0, eig_vals.shape[0] + 1), cum_var_exp, where="post", linewidth=3,
                         label="%s cumulative explained variance (95%%: %s / 99%%: %s)" % (model, np.sum(cum_var_exp < .95), np.sum(cum_var_exp < .99)))
            plt.yticks([0.0, .25, .50, .75, .95, .99])
            plt.ylabel("Explained variance ratio", fontsize=16)
            plt.xlabel("Principal components", fontsize=16)
            plt.legend(loc="best", fontsize=16, frameon=True)
            plt.tight_layout()
            plt.show()

    def plot_active_latent_units(
            self,
            organ,
            topology_version,
            cvs=None
    ):
        """
        Plots latent unit activity measured by empirical variance of the expected latent space.
        See: https://arxiv.org/abs/1509.00519
        If an embedding file is found that contains z, z_mean, z_var (eg. output from predict_variational() function)
        the model will use z, and not z_mean.
        """

        colors = ['red', 'blue', 'green', 'cyan', 'magenta', 'yellow', 'darkgreen', 'lime', 'navy', 'royalblue',
                  'pink', 'peru']

        def active_latent_units_mask(z):
            var_x = np.diagonal(np.cov(z.T))
            min_var_x = 0.01
            active_units_mask = var_x > min_var_x
            return active_units_mask

        import matplotlib.pyplot as plt
        if self.summary_tab is None:
            self.create_summary_tab()
        models = np.unique(self.summary_tab["model_type"]).tolist()
        self.summary_tab["topology"] = [x.split("_")[5] for x in self.summary_tab["model_gs_id"].values]

        with plt.style.context("seaborn-whitegrid"):
            plt.figure(figsize=(12, 6))
            plt.axhline(np.log(0.01), color="k", linestyle='dashed', linewidth=2, label="active unit threshold")
            for i, model in enumerate(models):
                model_id, embedding, covar = self.best_model_embedding(
                    subset={"model_type": model, "organ": organ, "topology": topology_version},
                    partition="val",
                    metric="loss",
                    cvs=cvs,
                )
                if len(embedding[0].shape) == 3:
                    z = embedding[0][0]  # in case of three-dimensional VAE embedding (z, z_mean, z_var), use z
                else:
                    z = embedding[0]
                latent_dim = z.shape[1]
                var = np.sort(np.diagonal(np.cov(z.T)))[::-1]
                log_var = np.log(var)
                active_units = np.log(var[active_latent_units_mask(z)])

                plt.plot(range(1, log_var.shape[0] + 1), log_var, color=colors[i], alpha=1.0, linewidth=3,
                         label="%s active units: %i" % (model, len(active_units)))
                # to plot vertical lines
                log_var_cut = var.copy()
                log_var_cut[~active_latent_units_mask(z)] = 0
                log_var_cut = np.log(log_var_cut)
                num_active = np.argmax(log_var_cut)
                if num_active > 0:
                    plt.vlines(num_active, ymin=-.15, ymax=0.15, color=colors[i], linestyle='solid', linewidth=3)
                if model == "vaevamp":
                    z1, z2 = np.split(np.log(np.diagonal(np.cov(z.T))), 2)
                    plt.plot(range(1, int(latent_dim / 2) + 1), np.sort(z2)[::-1], color=colors[i], alpha=1.0,
                             label=r"%s $z_2$ active units: %i" % (model, len(z2[z2 > np.log(0.01)])),
                             linestyle='dashed', linewidth=3)
                    plt.plot(range(1, int(latent_dim / 2) + 1), np.sort(z1)[::-1], color=colors[i], alpha=1.0,
                             label=r"%s $z_1$ active units: %i" % (model, len(z1[z1 > np.log(0.01)])),
                             linestyle='dotted', linewidth=3)
            plt.xlabel(r'Latent unit $i$', fontsize=16)
            plt.ylabel(r'$\log\,{(A_{\bf z})}_i$', fontsize=16)
            plt.title(r"Latent unit activity", fontsize=16)
            plt.legend(loc="upper right", frameon=True, fontsize=12)
            plt.tight_layout()
            plt.show()



import numpy as np
try:
    import tensorflow as tf
except ImportError:
    tf = None
from typing import Union, Tuple

from sfaira.models.embedding.output_layers import NegBinOutput, NegBinSharedDispOutput, NegBinConstDispOutput, \
    GaussianOutput, GaussianSharedStdOutput, GaussianConstStdOutput
from sfaira.versions.topologies import TopologyContainer
from sfaira.models.embedding.base import BasicModelKerasEmbedding
from sfaira.models.pp_layer import PreprocInput
from sfaira.models.made import MaskingDense


class Sampling(tf.keras.layers.Layer):
    """Uses (z_mean, z_log_var) to sample z."""

    def call(self, inputs, **kwargs):
        z_mean, z_log_var = inputs
        batch = tf.shape(z_mean)[0]
        dim = tf.shape(z_mean)[1]
        epsilon = tf.keras.backend.random_normal(shape=(batch, dim))
        return z_mean + tf.exp(0.5 * z_log_var) * epsilon


class Encoder(tf.keras.layers.Layer):
    """Maps input to embedding space"""

    def __init__(
            self,
            latent_dim: Tuple,
            dropout_rate,
            l1_coef: float,
            l2_coef: float,
            batchnorm: bool,
            activation: str,
            kernel_initializer: str,
            name='encoder',
            **kwargs
    ):
        super().__init__(name=name, **kwargs)

        self.fwd_pass = []
        for i, hid_size in enumerate(latent_dim[:-1]):
            self.fwd_pass.append(
                tf.keras.layers.Dense(
                    units=hid_size,
                    activation=activation,
                    kernel_initializer=kernel_initializer,
                    kernel_regularizer=tf.keras.regularizers.l1_l2(l1=l1_coef, l2=l2_coef),
                    name='enc%s' % i
                )
            )
            if batchnorm:
                self.fwd_pass.append(
                    tf.keras.layers.BatchNormalization(
                        center=True,
                        scale=True
                    )
                )
            self.fwd_pass.append(
                tf.keras.layers.Dropout(
                    dropout_rate,
                    noise_shape=None,
                    seed=None
                )
            )

        # final layer
        self.dense_mean = tf.keras.layers.Dense(
            units=latent_dim[-1],
            activation="linear",
            name='z_mean'
        )
        self.dense_log_sigma = tf.keras.layers.Dense(
            units=latent_dim[-1],
            activation="linear",
            name='z_log_var'
        )
        # IAF layers
        self.dense_h = tf.keras.layers.Dense(
            units=latent_dim[-1],
            activation='sigmoid',
            name='h'
        )

    def call(self, inputs, **kwargs):
        x = inputs
        for layer in self.fwd_pass:
            x = layer(x)
        # final layer
        z_mean = self.dense_mean(x)
        z_log_var = self.dense_log_sigma(x)
        h = self.dense_h(x)

        return z_mean, z_log_var, h


class IAF(tf.keras.layers.Layer):
    def __init__(
        self,
        bottleneck: int,
        n_iaf: int,
        l1_coef: float,
        l2_coef: float,
        masking_dim=320,
        n_made=2,
        activation="relu",
        name='iaf',
        **kwargs
    ):
        """
        Transforms latent space with simple distribution to one with a more flexible one.

        :param latent_dim: dimension of bottleneck
        :param masking_dim: units of autoregressive networks
        :param n_made: number of autoregressive networks
        :param n_iaf: number of IAF transforms
        :param l1_coef:
        :param l2_coef:
        :param activation:
        :param name:
        :param kwargs:
        """
        super().__init__(name=name, **kwargs)
        self.m_fwd_pass = []
        self.s_fwd_pass = []
        for i in range(n_iaf):
            self.m_fwd_pass.append(
                MaskingDense(
                    units=masking_dim,
                    out_units=bottleneck,
                    hidden_layers=n_made,
                    activation=activation,
                    out_activation='linear',
                    # batchnorm=True,
                    name='mean_%s' % i)
            )
            self.s_fwd_pass.append(
                MaskingDense(
                    units=masking_dim,
                    out_units=bottleneck,
                    hidden_layers=n_made,
                    kernel_regularizer=tf.keras.regularizers.l1_l2(l1=l1_coef, l2=l2_coef),
                    out_bias_initializer='ones',
                    activation=activation,
                    out_activation='sigmoid',
                    # batchnorm=True,
                    name='sigma_%s' % i)
            )

    def call(self, inputs, **kwargs):
        """
        :param inputs: initial latent space z and dense layer h
        :param kwargs:
        :return: z_t: transformed latent space
        :return: s_t_sigmas: list of applied transformations
        """
        z_t, h = inputs

        s_t_sigmas = []
        for i, (m_iaf, s_iaf) in enumerate(zip(self.m_fwd_pass, self.s_fwd_pass)):
            if i > 0:
                z_t = tf.keras.layers.Lambda(lambda x: tf.keras.backend.reverse(x, axes=-1))(z_t)
                # z_t = tf.gather(z_t, tf.random.shuffle(tf.range(tf.shape(z_t)[0])))  # random permutation
            m_t = m_iaf([z_t, h])
            sigma_t = s_iaf([z_t, h])
            z_t = z_t * sigma_t + m_t * (1. - sigma_t)
            s_t_sigmas.append(sigma_t)
        return z_t, s_t_sigmas


class Decoder(tf.keras.layers.Layer):
    """Maps latent space sample back to output"""

    def __init__(
            self,
            latent_dim: Tuple,
            dropout_rate,
            l1_coef: float,
            l2_coef: float,
            batchnorm: bool,
            activation: str,
            kernel_initializer: str,
            name='decoder',
            **kwargs
    ):
        super().__init__(name=name, **kwargs)

        self.fwd_pass = []
        for i, hid_size in enumerate(latent_dim):
            self.fwd_pass.append(
                tf.keras.layers.Dense(
                    units=hid_size,
                    activation=activation,
                    kernel_initializer=kernel_initializer,
                    kernel_regularizer=tf.keras.regularizers.l1_l2(l1=l1_coef, l2=l2_coef),
                    name='dec%s' % i
                )
            )
            if batchnorm:
                self.fwd_pass.append(
                    tf.keras.layers.BatchNormalization(
                        center=True,
                        scale=True
                    )
                )
            self.fwd_pass.append(
                tf.keras.layers.Dropout(
                    dropout_rate,
                    noise_shape=None,
                    seed=None
                )
            )

    def call(self, inputs, **kwargs):
        x = inputs
        for layer in self.fwd_pass:
            x = layer(x)
        return x


class ModelKerasVaeIAF(BasicModelKerasEmbedding):

    def __init__(
            self,
            in_dim,
            latent_dim=(128, 64, 2, 64, 128),
            n_iaf=2,
            dropout_rate=0.1,
            l2_coef: float = 0.,
            l1_coef: float = 0.,
            mc_samples=10,
            batchnorm=False,
            activation='tanh',
            init='glorot_uniform',
            output_layer="nb"
    ):
        super(ModelKerasVaeIAF, self).__init__()
        # Check length of latent dim to divide encoder-decoder stack:
        if len(latent_dim) % 2 == 1:
            n_layers_enc = len(latent_dim) // 2 + 1
        else:
            raise ValueError("len(latent_dim)=%i should be uneven to provide a defined bottleneck" % len(latent_dim))

        inputs_encoder = tf.keras.Input(shape=(in_dim,), name='counts')
        inputs_sf = tf.keras.Input(shape=(1,), name='size_factors')
        inputs_encoder_pp = PreprocInput()(inputs_encoder)
        encoder = Encoder(
            latent_dim=latent_dim[:n_layers_enc],
            dropout_rate=dropout_rate,
            l1_coef=l1_coef,
            l2_coef=l2_coef,
            batchnorm=batchnorm,
            activation=activation,
            kernel_initializer=init
        )
        iaf = IAF(
            bottleneck=latent_dim[n_layers_enc - 1],
            n_iaf=n_iaf,
            l1_coef=l1_coef,
            l2_coef=l2_coef
        )
        decoder = Decoder(
            latent_dim=latent_dim[n_layers_enc:],
            dropout_rate=dropout_rate,
            l1_coef=l1_coef,
            l2_coef=l2_coef,
            batchnorm=batchnorm,
            activation=activation,
            kernel_initializer=init
        )
        sampling = Sampling()
        z_mean, z_log_var, h = encoder(inputs_encoder_pp)

        # for prediction
        z_0 = sampling([z_mean, z_log_var])
        z_t, s_t_sigmas = iaf([z_0, h])
        z_t_square_mc = 0
        z_t_mean = 0

        # for kl divergence
        for i in range(mc_samples):
            z = sampling([z_mean, z_log_var])
            z, s_t_sigmas = iaf([z, h])
            z_t_square_mc += tf.square(z)
            z_t_mean += z
        z_t_square_mc = z_t_square_mc / mc_samples
        z_t_mean = z_t_mean / mc_samples

        cum_s_t_log_var = 0
        for s_t_sigma in s_t_sigmas:
            cum_s_t_log_var += 2.0 * tf.keras.backend.log(s_t_sigma + 1e-10)

        expected_logqz_x = -0.5 * tf.reduce_sum(1 + z_log_var + cum_s_t_log_var, axis=1)
        expected_logpz = -0.5 * tf.reduce_sum(z_t_square_mc, axis=1)

        expected_densities = tf.keras.layers.Concatenate(axis=1, name='kl')([
            tf.expand_dims(expected_logqz_x, axis=1),
            tf.expand_dims(expected_logpz, axis=1)])

        output_decoder = decoder(z_t)

        if output_layer == 'nb':
            output_decoder_expfamily = NegBinOutput(original_dim=in_dim)((output_decoder, inputs_sf))
        elif output_layer == 'nb_shared_disp':
            output_decoder_expfamily = NegBinSharedDispOutput(original_dim=in_dim)((output_decoder, inputs_sf))
        elif output_layer == 'nb_const_disp':
            output_decoder_expfamily = NegBinConstDispOutput(original_dim=in_dim)((output_decoder, inputs_sf))
        elif output_layer == 'gaussian':
            output_decoder_expfamily = GaussianOutput(original_dim=in_dim)((output_decoder, inputs_sf))
        elif output_layer == 'gaussian_shared_disp':
            output_decoder_expfamily = GaussianSharedStdOutput(original_dim=in_dim)((output_decoder, inputs_sf))
        elif output_layer == 'gaussian_const_disp':
            output_decoder_expfamily = GaussianConstStdOutput(original_dim=in_dim)((output_decoder, inputs_sf))
        else:
            raise ValueError("tried to access a non-supported output layer %s" % output_layer)
        output_decoder_expfamily_concat = tf.keras.layers.Concatenate(axis=1, name="neg_ll")(output_decoder_expfamily)

        self.encoder_model = tf.keras.Model(
            inputs=[inputs_encoder, inputs_sf],
            outputs=[z_t, z_t_mean, z_0],
            name="encoder_model"
        )
        self.training_model = tf.keras.Model(
            inputs=[inputs_encoder, inputs_sf],
            outputs=[output_decoder_expfamily_concat, expected_densities],
            name="autoencoder"
        )

    def predict_embedding(self, x, variational=False, return_z0=False):
        if return_z0 and variational:
            z_t, z_t_mean, z_0 = self.encoder_model.predict(x)
            return z_t, z_t_mean, z_0
        elif not return_z0 and variational:
            z_t, z_t_mean = self.encoder_model.predict(x)[:2]
            return z_t, z_t_mean
        elif return_z0 and not variational:
            z_t_mean, z_0 = self.encoder_model.predict(x)[1:]
            return z_t_mean, z_0
        elif not return_z0 and not variational:
            z_t_mean = self.encoder_model.predict(x)[1]
            return z_t_mean


class ModelVaeIAFVersioned(ModelKerasVaeIAF):
    def __init__(
            self,
            topology_container: TopologyContainer,
            override_hyperpar: Union[dict, None] = None
    ):
        hyperpar = topology_container.topology["hyper_parameters"]
        if override_hyperpar is not None:
            for k in list(override_hyperpar.keys()):
                hyperpar[k] = override_hyperpar[k]
        super().__init__(
            in_dim=topology_container.n_var,
            **hyperpar
        )
        print('passed hyperpar: \n', hyperpar)
        self._topology_id = topology_container.topology_id
        self.genome_size = topology_container.n_var
        self.model_class = "embedding"
        self.model_type = topology_container.model_type
        self.hyperparam = dict(
            list(hyperpar.items()) +  # noqa: W504
            [
                ("topology_id", self._topology_id),
                ("genome_size", self.genome_size),
                ("model_class", self.model_class),
                ("model_type", self.model_type)
            ]
        )



import numpy as np
try:
    import tensorflow as tf
except ImportError:
    tf = None
from typing import List, Union

from sfaira.versions.metadata import CelltypeUniverse
from sfaira.versions.topologies import TopologyContainer
from sfaira.models.celltype.base import BasicModelKerasCelltype
from sfaira.models.pp_layer import PreprocInput


class CellTypeMlp(BasicModelKerasCelltype):
    """
    Multi-layer perceptron to predict cell type.

    """

    def __init__(
            self,
            in_dim: int,
            out_dim: int,
            units: List[int] = [],
            activation="relu",
            use_bias=True,
            l1_coef: float = 0.,
            l2_coef: float = 0.,
            kernel_initializer='glorot_uniform',
            bias_initializer='zeros',
            bias_regularizer=None,
            activity_regularizer=None,
            kernel_constraint=None,
            bias_constraint=None,
            name='celltype_mlp',
            **kwargs
    ):
        """

        :param in_dim: Input feature dimension (number of features).
        :param out_dim: Number of output nodes: Number of cell types +1 for unkown cell type.
        :param units: See tf.keras.layers.Dense() documentation.
        :param activation: See tf.keras.layers.Dense() documentation.
        :param use_bias: See tf.keras.layers.Dense() documentation.
        :param kernel_initializer: See tf.keras.layers.Dense() documentation.
        :param bias_initializer: See tf.keras.layers.Dense() documentation.
        :param kernel_regularizer: See tf.keras.layers.Dense() documentation.
        :param bias_regularizer: See tf.keras.layers.Dense() documentation.
        :param activity_regularizer: See tf.keras.layers.Dense() documentation.
        :param kernel_constraint: See tf.keras.layers.Dense() documentation.
        :param bias_constraint: See tf.keras.layers.Dense() documentation.
        """
        self.in_dim = in_dim
        self.out_dim = out_dim

        inputs = tf.keras.Input(shape=(in_dim,))
        x = PreprocInput()(inputs)
        depth = len(units) + 1
        for i in range(depth):
            x = tf.keras.layers.Dense(
                units=units[i] if i < (depth - 1) else out_dim,
                activation=activation if i < (depth - 1) else "softmax",
                use_bias=use_bias,
                kernel_initializer=kernel_initializer,
                bias_initializer=bias_initializer,
                kernel_regularizer=tf.keras.regularizers.l1_l2(l1=l1_coef, l2=l2_coef),
                bias_regularizer=bias_regularizer,
                activity_regularizer=activity_regularizer,
                kernel_constraint=kernel_constraint,
                bias_constraint=bias_constraint
            )(x)
        self.training_model = tf.keras.Model(inputs=[inputs, ], outputs=x, name=name)


class CellTypeMlpVersioned(CellTypeMlp):

    def __init__(
            self,
            celltypes_version: CelltypeUniverse,
            topology_container: TopologyContainer,
            override_hyperpar: Union[dict, None] = None
    ):
        """

        :param celltypes_version:
        :param topology_container:
        :param override_hyperpar: Dictionary with hyper-parameters of model to override in preset hyper-parameter
            dictionary that is queried based on the topology_id. Can contain a subset of all hyperparameters.
        """
        # Get cell type version instance based on topology ID, organism and organ.
        hyperpar = topology_container.topology["hyper_parameters"]
        if override_hyperpar is not None:
            for k in list(override_hyperpar.keys()):
                hyperpar[k] = override_hyperpar[k]
        super().__init__(
            in_dim=topology_container.n_var,
            out_dim=celltypes_version.onto_cl.n_leaves,
            **hyperpar
        )
        print('passed hyperpar: \n', hyperpar)
        self._topology_id = topology_container.topology_id
        self.genome_size = topology_container.n_var
        self.model_class = "celltype"
        self.model_type = topology_container.model_type
        self.hyperparam = dict(
            list(hyperpar.items()) +  # noqa: W504
            [
                ("topology_id", self._topology_id),
                ("genome_size", self.genome_size),
                ("model_class", self.model_class),
                ("model_type", self.model_type),
                ("ntypes", celltypes_version.onto_cl.n_leaves),
            ]
        )



import shutil
import time

from sfaira import settings
from sfaira.consts import AdataIdsSfaira, OCS


class CacheControl:

    def __init__(self):
        self.adata_ids = AdataIdsSfaira()

    @staticmethod
    def clear():
        for i in range(5):
            print(f"WARNING: deleting entire folder {settings.cachedir_ontologies} in {5 - i} sec ...")
            time.sleep(1)
        shutil.rmtree(path=settings.cachedir_ontologies)

    def reload(self):
        for x in self.adata_ids.ontology_constrained:
            OCS.reload_ontology(attr=x)



from random import shuffle
from typing import List, Tuple

import numpy as np
from numba import njit


def _split_idx_along_arr_chunks(indicies: np.ndarray, chunks: Tuple[int]):
    assert np.all(np.diff(indicies) >= 0.)  # make sure array is sorted
    assert indicies.dtype == np.int64

    @njit
    def split_along_array_boundaries(idxs: np.ndarray, arr_boundaries: np.ndarray):
        idx_split = []  # variable to accumulate index groups

        partition_ctr = 0
        start_next_partition = arr_boundaries[partition_ctr]
        idxs_partition = [np.int64(x) for x in range(0)]  # temporary variable to accumulate idxs for a index group
        for idx in idxs:
            if idx < start_next_partition:
                # append idx to index group
                idxs_partition.append(idx)
            else:
                # find next bigger partition
                while idx >= start_next_partition:
                    partition_ctr += 1
                    start_next_partition = arr_boundaries[partition_ctr]
                # start a new index group list
                if len(idxs_partition) > 0:
                    idx_split.append(idxs_partition)
                    idxs_partition = [np.int64(x) for x in range(0)]
                # append idx to index group
                idxs_partition.append(idx)
        if len(idxs_partition) > 0:
            idx_split.append(idxs_partition)

        return idx_split

    return [np.array(xx) for xx in split_along_array_boundaries(indicies, np.cumsum(chunks))]


class BatchDesignBase:

    """
    Manages distribution of selected indices for a given data object over subsequent batches.

    This involves randomisation and possible meta-data dependent batching.
    This class is centred on the property `.design` which yields a list of observation indices (arrays), where each
    array is the set of indices that map to a batch and the sequence of batches in the list encodes the sequence of
    batches during an epoch over the data.
    """

    def __init__(self,
                 retrieval_batch_size: int,
                 randomized_batch_access: bool,
                 random_access: bool,
                 **kwargs):
        self.retrieval_batch_size = retrieval_batch_size
        self._batches = None
        self._idx = None
        self._batch_splits = None
        if randomized_batch_access and random_access:
            raise ValueError("Do not use randomized_batch_access and random_access.")
        self.randomized_batch_access = randomized_batch_access
        self.random_access = random_access

    @property
    def batchsplits(self) -> Tuple[int]:
        """
        Batch size of the yieleded batches.
        """
        return self._batch_splits

    @batchsplits.setter
    def batchsplits(self, batch_splits: Tuple[int]):
        self._batch_splits = batch_splits

    @property
    def n_batches(self) -> int:
        return len(self.design)

    @property
    def idx(self):
        """
        Protects property from uncontrolled changing.
        Changes to _idx require changes to _batch_bounds.
        """
        return self._idx

    @idx.setter
    def idx(self, x):
        self._idx = np.sort(x)  # idx has to be sorted for logic in subclasses

    @property
    def design(self) -> List[np.array]:
        """
        Yields index objects for one epoch of all data.

        These index objects are used by generators that have access to the data objects to build data batches.
        Randomization is performed anew with every call to this property.

        :returns: List[np.array]
            List of indices per batch
        """
        raise NotImplementedError()


class BatchDesignBasic(BatchDesignBase):

    """Standard batched access to data."""

    @property
    def design(self) -> List[np.ndarray]:

        if self.random_access:
            # shuffle idx for random access
            idx = np.random.permutation(self.idx)
            batches = np.array_split(idx, max(len(idx) // self.retrieval_batch_size, 1))
        elif self.randomized_batch_access:
            if self.batchsplits is not None:
                # if zarr chunk sizes are known -> split first along those and shuffle those batches
                batches = _split_idx_along_arr_chunks(self.idx, self.batchsplits)
                shuffle(batches)
                # accumulate smaller batches to the size of retrieval_batch_size
                batches = np.array_split(np.concatenate(batches), max(len(self.idx) // self.retrieval_batch_size, 1))
            else:
                batches = np.array_split(self.idx, max(len(self.idx) // self.retrieval_batch_size, 1))
                shuffle(batches)
        else:
            batches = np.array_split(self.idx, max(len(self.idx) // self.retrieval_batch_size, 1))

        return batches


class BatchDesignBalanced(BatchDesignBase):

    """
    Balanced batches across meta data partitions of data.
    """

    def __init__(self, grouping, group_weights: dict, randomized_batch_access: bool, random_access: bool,
                 **kwargs):
        """

        :param grouping: Group label for each entry in idx.
        :param group_weights: Group weight for each unique group in grouping. Does not have to normalise to a probability
            distribution but is normalised in this function. The outcome vector is always of length idx.
        """
        super(BatchDesignBalanced, self).__init__(randomized_batch_access=randomized_batch_access,
                                                  random_access=random_access, **kwargs)
        if randomized_batch_access:
            print("WARNING: randomized_batch_access==True is not a meaningful setting for BatchDesignBalanced. "
                  "Setting will be ignored!p")
        if not random_access:
            print("WARNING: random_access==False is dangerous if you do not work with a large shuffle buffer "
                  "downstream of the sfaira generator.")
        # Create integer group assignment array.
        groups = np.sort(list(group_weights.keys()))
        grouping_int = np.zeros((grouping.shape[0], ), dtype="int32") - 1
        for i, x in enumerate(groups):
            grouping_int[np.where(grouping == x)[0]] = i
        assert np.all(grouping_int >= 0)
        # Create sampling weights: Sampling weights are a probability distribution over groups.
        weight_group = np.array([group_weights[x] for x in groups])
        p_obs = np.asarray(weight_group[grouping_int], dtype="float64")
        p_obs = p_obs / np.sum(p_obs)
        if np.any(p_obs == 0.):
            raise ValueError(f"Down-sampling resulted in zero-probability weights on cells. "
                             f"Group weights: {weight_group}")
        self.p_obs = p_obs

    @property
    def design(self) -> List[np.ndarray]:
        # select relevant probabilities and renormalize to prob vector
        p_obs = self.p_obs[self.idx]
        p_obs = p_obs / np.sum(p_obs)
        # Re-sample index vector.
        idx = np.random.choice(a=self.idx, replace=True, size=len(self.idx), p=p_obs)
        if not self.random_access:  # Note: randomization is result from sampling above, need to revert if not desired.
            idx = np.sort(idx)
        batches = np.array_split(idx, max(len(idx) // self.retrieval_batch_size, 1))

        return batches


class BatchDesignBlocks(BatchDesignBase):

    """
    Meta data-defined blocks of observations in each batch.

    Note that the batches do not necessarily contain equal numbers of observations! The yielded batch size depends
    solely on the number of observations in a meta data partition.
    """

    def __init__(self, grouping, random_access: bool, **kwargs):
        """

        :param grouping: List of group labels for each element in idx. Must have same length as data array.
        :param group_weights: Group weight for each unique group in grouping. Does not have to normalise to a probability
            distribution but is normalised in this function. The outcome vector is always of length idx.
        """
        super(BatchDesignBlocks, self).__init__(random_access=random_access, **kwargs)
        if not random_access:
            print("WARNING: random_access==False is dangerous if you do not work with a large shuffle buffer "
                  "downstream of the sfaira generator.")
        # Create integer group assignment array.
        self.grouping = grouping

    @property
    def grouping(self):
        return self._grouping

    @grouping.setter
    def grouping(self, x):
        self._grouping = x
        # Reset:
        self._groups = None
        self.idx_sorted = None

    @property
    def groups(self):
        if self._groups is None:
            self._groups = np.unique(self.grouping)
        return self._groups

    @property
    def idx(self):
        """
        Protects property from uncontrolled changing.
        Changes to _idx require changes to batch splitting.
        """
        return self._idx

    @idx.setter
    def idx(self, x):
        self._idx = x
        # Reset:
        self._groups = None
        idx_sorted = []
        for group in self.groups:
            # subselect by group and only keep indicies which are in the supplied index range
            idx = np.intersect1d(np.where(self.grouping == group)[0], x)
            idx_sorted.append(idx)
        self.idx_sorted = idx_sorted

    @property
    def design(self) -> List[np.ndarray]:
        idx_sorted = self.idx_sorted
        batches = []
        for idxs_group in idx_sorted:
            if len(idxs_group) > 0:
                if self.random_access:
                    # shuffle subgroups if random_access
                    idxs_group = np.random.permutation(idx_sorted)
                batches.append(idxs_group)
        if self.random_access:
            # shuffle blocks
            shuffle(batches)

        return batches


class BatchDesignFull(BatchDesignBase):

    """Emits full dataset as a single batch in each query."""

    @property
    def design(self) -> List[np.ndarray]:
        idx = np.arange(0, len(self.idx))
        if self.random_access:
            # shuffle idx for random access
            idx = np.random.permutation(idx)
        return [idx]


BATCH_SCHEDULE = {
    "base": BatchDesignBasic,
    "balanced": BatchDesignBalanced,
    "blocks": BatchDesignBlocks,
    "full": BatchDesignFull,
}



import numpy as np
try:
    import tensorflow as tf
except ImportError:
    tf = None
from typing import List, Union

from sfaira.models.embedding.output_layers import NegBinOutput, NegBinSharedDispOutput, NegBinConstDispOutput, \
    GaussianOutput, GaussianSharedStdOutput, GaussianConstStdOutput
from sfaira.versions.topologies import TopologyContainer
from sfaira.models.embedding.base import BasicModelKerasEmbedding
from sfaira.models.pp_layer import PreprocInput


class EncoderLinear(tf.keras.layers.Layer):
    """Maps input to embedding space"""

    def __init__(
            self,
            latent_dim: int,
            positive_components: bool,
            l1_coef: float,
            l2_coef: float,
            name='encoder',
            **kwargs
    ):
        super().__init__(name=name, **kwargs)
        self.fwd_pass = []
        self.fwd_pass.append(tf.keras.layers.Dense(
            units=latent_dim,
            activation="linear" if not positive_components else "relu",
            kernel_regularizer=tf.keras.regularizers.l1_l2(l1=l1_coef, l2=l2_coef)
        ))

    def call(self, inputs, **kwargs):
        x = inputs
        for layer in self.fwd_pass:
            x = layer(x)
        return x


class ModelKerasLinear(BasicModelKerasEmbedding):

    def __init__(
            self,
            in_dim,
            latent_dim: int = 10,
            positive_components: bool = False,
            l2_coef: float = 0.,
            l1_coef: float = 0.,
            dropout_rate=None,
            output_layer="nb"
    ):
        super(ModelKerasLinear, self).__init__()

        self.in_dim = in_dim
        self.latent_dim = latent_dim
        out_dim = in_dim
        self.out_dim = in_dim

        inputs_encoder = tf.keras.Input(shape=(in_dim,), name='counts')
        inputs_sf = tf.keras.Input(shape=(1,), name='size_factors')
        inputs_encoder_pp = PreprocInput()(inputs_encoder)
        output_encoder = EncoderLinear(
            latent_dim=latent_dim,
            positive_components=positive_components,
            l1_coef=l1_coef,
            l2_coef=l2_coef
        )(inputs_encoder_pp)

        if output_layer == 'nb':
            output_decoder_expfamily = NegBinOutput(original_dim=out_dim)((output_encoder, inputs_sf))
        elif output_layer == 'nb_shared_disp':
            output_decoder_expfamily = NegBinSharedDispOutput(original_dim=out_dim)((output_encoder, inputs_sf))
        elif output_layer == 'nb_const_disp':
            output_decoder_expfamily = NegBinConstDispOutput(original_dim=out_dim)((output_encoder, inputs_sf))
        elif output_layer == 'gaussian':
            output_decoder_expfamily = GaussianOutput(original_dim=out_dim)((output_encoder, inputs_sf))
        elif output_layer == 'gaussian_shared_disp':
            output_decoder_expfamily = GaussianSharedStdOutput(original_dim=out_dim)((output_encoder, inputs_sf))
        elif output_layer == 'gaussian_const_disp':
            output_decoder_expfamily = GaussianConstStdOutput(original_dim=out_dim)((output_encoder, inputs_sf))
        else:
            raise ValueError("tried to access a non-supported output layer %s" % output_layer)
        output_decoder_expfamily_concat = tf.keras.layers.Concatenate(axis=1, name="neg_ll")(output_decoder_expfamily)

        self.encoder_model = tf.keras.Model(
            inputs=[inputs_encoder, inputs_sf],
            outputs=output_encoder,
            name="encoder"
        )
        self.training_model = tf.keras.Model(
            inputs=[inputs_encoder, inputs_sf],
            outputs=[output_decoder_expfamily_concat],
            name="autoencoder"
        )


class ModelLinearVersioned(ModelKerasLinear):
    def __init__(
            self,
            topology_container: TopologyContainer,
            override_hyperpar: Union[dict, None] = None
    ):
        hyperpar = topology_container.topology["hyper_parameters"]
        if override_hyperpar is not None:
            for k in list(override_hyperpar.keys()):
                hyperpar[k] = override_hyperpar[k]
        super().__init__(
            in_dim=topology_container.n_var,
            **hyperpar
        )
        print('passed hyperpar: \n', hyperpar)
        self._topology_id = topology_container.topology_id
        self.genome_size = topology_container.n_var
        self.model_class = "embedding"
        self.model_type = topology_container.model_type
        self.hyperparam = dict(
            list(hyperpar.items()) +  # noqa: W504
            [
                ("topology_id", self._topology_id),
                ("genome_size", self.genome_size),
                ("model_class", self.model_class),
                ("model_type", self.model_type)
            ]
        )



import anndata
import os
import scipy.sparse
import pandas as pd
import gzip
import scipy.io


def load(data_dir, **kwargs):
    fn = [os.path.join(data_dir, "GSE158055_covid19_counts.mtx.gz"),
          os.path.join(data_dir, "GSE158055_covid19_barcodes.tsv.gz"),
          os.path.join(data_dir, "GSE158055_covid19_features.tsv.gz")]
    fn_meta = [os.path.join(data_dir, "GSE158055_cell_annotation.csv.gz"),
               os.path.join(data_dir, "GSE158055_sample_metadata.xlsx")]

    # Data is not in sc.read_10x_mtx format because GSE158055_covid19_counts does not end on "_matrix".
    with gzip.open(fn[0], 'rb') as mm:
        X = scipy.io.mmread(mm).T.tocsr()
    obs = pd.read_csv(fn[1], header=None, sep="\t", index_col=0)
    obs.index.name = None
    var = pd.read_csv(fn[2], header=None, sep="\t", names=['names'])
    var.index = var['names'].values
    adata = anndata.AnnData(X=X, obs=obs, var=var)

    annotation = pd.read_csv(fn_meta[0])
    metadata = pd.read_excel(fn_meta[1], skiprows=20, skipfooter=37)
    metadata.columns = metadata.columns.str.split('characteristics: ').str[-1]
    tmp = annotation.merge(metadata, left_on='sampleID', right_on='Sample name').set_index('cellName')
    adata.obs = adata.obs.join(tmp).astype(str)

    return adata



from pytest import fixture


def pytest_addoption(parser):
    parser.addoption(
        "--doi_sfaira_repr",
        action="store"
    )
    parser.addoption(
        "--test_data",
        action="store"
    )


@fixture()
def doi_sfaira_repr(request):
    return request.config.getoption("--doi_sfaira_repr")


@fixture()
def test_data(request):
    return request.config.getoption("--test_data")



import abc
try:
    import tensorflow as tf
except ImportError:
    tf = None


class BasicModel(abc.ABC):
    """
    This base class defines model attributes shared across all models.
    """
    _version: str
    _topology_id: str
    genome_size: int
    hyperparam: dict
    model_class: str
    model_type: str

    @property
    def version(self):
        return self._version


class BasicModelKeras(BasicModel):
    """
    This base class defines model attributes shared across all tf.keras models.
    """

    training_model: tf.keras.Model



from functools import partial

import pandas as pd

from sfaira.data.store.batch_schedule import BatchDesignBase


class CartBase:
    """
    A cart manages data emission for a specified subset of a store.

    Data can be queried through:

        1) an iterator function
        2) the full data objects (.x and .obs)

    1) Iterator function:
    The iterator can often directly be used without a class like this one around it.
    However, that often implies that the namespace with the pointer to the data set is destroyed after iterator
    function declaration, which means that the pointer needs to be redefined for every full pass over the iterator.
    The class around this property maintains the namespace that includes the pointer and its instance can be used to
    avoid redefining the pointer every time the generator runs out and is re-called.
    For this run time advantage to not be compromised by memory load and class initialisation run time cost induced by
    actually copying data objects, it is important that the data object stored in this class is indeed a pointer.
    This is the case for:

        - lazily loaded dask arrays
        - anndata.Anndata view

    which have their own classes below.
    """

    schedule: BatchDesignBase
    var: pd.DataFrame  # Feature meta data (features x properties).
    map_fn: callable

    def adaptor(
            self,
            generator_type: str,
            dataset_kwargs: dict = None,
            shuffle_buffer: int = 0,
            repeat: int = 1,
            **kwargs
    ):
        """
        The adaptor turns a python base generator into a different iteratable object, defined by generator_type.

        :param generator_type: Type of output iteratable.
            - python base generator (no change to `.generator`)
            - tensorflow dataset: This dataset is defined on a python iterator.
                Important:
                    This only returns the tf.data.Dataset.from_generator(). You need to define the input pipeline
                    (e.g. .batch(), .prefetch()) on top of this data set.
            - pytorch: We distinguish torch.data.Dataset and torch.data.DataLoader ontop of either.
                The Dataset vs DataLoader distinction is made by the "" suffix for Dataset or "-loader" suffix for +
                dataloader. The distinction between Dataset and IteratableDataset defines if the object is defined
                directly on a dask array or based on a python iterator on a dask array. Note that the python iterator
                can implement favorable remote access schemata but the torch.data.Dataset generally causes less trouble
                in out-of-the-box usage.
                    - torch.data.Dataset: "torch" prefix, ie "torch" or "torch-loader"
                    - torch.data.IteratableDataset: "torch-iter" prefix, ie "torch-iter" or "torch-iter-loader"
                Important:
                    For model training in pytorch you need the "-loader" prefix. You can specify the arguments passed
                    to torch.utils.data.DataLoader by the dataset_kwargs dictionary.
        :param dataset_kwargs: Dict
            Parameters to pass to the constructor of torch Dataset.
            Only relevant if generator_type in ['torch', 'torch-loader']
        :param shuffle_buffer: int
            If shuffle_buffer > 0 -> Use a shuffle buffer with size shuffle_buffer to shuffle output of self.iterator
            (this option is useful when using randomized_batch_access in the DaskCart)
        :param repeat: int
            Number of times to repeat the dataset until the underlying generator runs out of samples.
            If repeat <= 0 -> repeat dataset forever
        :returns: Modified iteratable (see generator_type).
        """
        if not dataset_kwargs:
            dataset_kwargs = {}

        if generator_type == "python":
            g = self.iterator(repeat=repeat, shuffle_buffer=shuffle_buffer)
        elif generator_type == "tensorflow":
            import tensorflow as tf

            g = tf.data.Dataset.from_generator(
                generator=partial(self.iterator, repeat=repeat, shuffle_buffer=shuffle_buffer), **kwargs
            )
        elif generator_type in ["torch", "torch-loader"]:
            from torch.utils.data import DataLoader
            # Only import this module if torch is used to avoid strict torch dependency:
            from sfaira.data.store.torch_dataset import SfairaDataset

            g = SfairaDataset(map_fn=self.map_fn, obs=self.obs, x=self.x, **dataset_kwargs)
            if generator_type == "torch-loader":
                g = DataLoader(g, **kwargs)
        elif generator_type in ["torch-iter", "torch-iter-loader"]:
            from torch.utils.data import DataLoader
            # Only import this module if torch is used to avoid strict torch dependency:
            from sfaira.data.store.torch_dataset import SfairaIterableDataset

            g = SfairaIterableDataset(iterator_fun=partial(self.iterator, repeat=repeat, shuffle_buffer=shuffle_buffer))
            if generator_type == "torch-iter-loader":
                g = DataLoader(g, **kwargs)
        else:
            raise ValueError(f"{generator_type} not recognized")
        return g

    @property
    def adata(self):
        """
        Assembles a slice of this cart based on .obs_idx as an anndata instance.
        """
        raise NotImplementedError()

    def iterator(self, repeat: int = 1, shuffle_buffer: int = 0):
        """
        Iterator over data matrix and meta data table, yields batches of data points.
        """
        raise NotImplementedError()

    @property
    def obs_idx(self):
        """
        Integer observation indices to select from cart. These will be emitted if data is queried.
        """
        raise NotImplementedError()

    @property
    def n_batches(self):
        return self.schedule.n_batches

    @property
    def n_obs(self):
        """Total number of observations in cart."""
        raise NotImplementedError()

    @property
    def n_obs_selected(self):
        """Total number of selected observations in cart."""
        raise NotImplementedError()

    @property
    def n_var(self):
        """Total number of features defined for return in cart."""
        raise NotImplementedError()

    @property
    def obs(self):
        """
        Selected meta data matrix (cells x meta data) that is emitted in batches by .iterator().
        """
        raise NotImplementedError()

    def move_to_memory(self):
        """
        Load underlying array into memory into memory.
        """
        raise NotImplementedError()

    @property
    def x(self):
        """
        Selected data matrix (cells x features) that is emitted in batches by .iterator().
        """
        raise NotImplementedError()



import anndata
from IPython.display import display_javascript, display_html
import json
import os
import pickle
import requests
from typing import List, Union
import uuid

from sfaira import settings
from sfaira.data.dataloaders.base import DatasetBase
from sfaira.consts import AdataIdsCellxgene, AdataIdsCellxgene_v2_0_0
from sfaira.data.dataloaders.databases.cellxgene.rest_helpers import get_collection, get_data
from sfaira.data.dataloaders.databases.cellxgene.rest_helpers import CELLXGENE_PRODUCTION_ENDPOINT, DOWNLOAD_DATASET


def cellxgene_fn(dir, dataset_id):
    return os.path.join(dir, dataset_id + ".h5ad")


def clean_cellxgene_meta_obs(k, val, adata_ids) -> Union[str, List[str]]:
    """
    :param k: Found meta data name.
    :param val: Found meta data entry.
    :returns: Cleaned meta data entry.
    """
    if k == adata_ids.organ:
        # Organ labels contain labels on tissue type also, such as 'UBERON:0001911 (cell culture)'.
        val = [v.split(" ")[0] for v in val]
    return val


def clean_cellxgene_meta_uns(k, val, adata_ids) -> Union[str, List[str]]:
    """
    :param k: Found meta data name.
    :param val: Found meta data entry.
    :returns: Cleaned meta data entry.
    """
    x_clean = []
    for v in val:
        # Decide if labels are read from name or ontology ID:
        v = v[adata_ids.onto_id_suffix[1:]]
        # Organ labels contain labels on tissue type also, such as 'UBERON:0001911 (cell culture)'.
        if k == adata_ids.organ:
            v = v.split(" ")[0]
        if v != adata_ids.unknown_metadata_identifier and v != adata_ids.invalid_metadata_identifier:
            x_clean.append(v)
    return x_clean


class Dataset(DatasetBase):
    """
    This is a dataloader for downloaded h5ad from cellxgene.

    In contrast to the base class, each instance is coupled to a particular collection_id to allow query.
    In the base classes, collection_id are only defined on the group level.
    """

    collection_id: str

    def __init__(
            self,
            collection_id: str = "default",
            data_path: Union[str, None] = None,
            meta_path: Union[str, None] = None,
            cache_path: Union[str, None] = None,
            load_func=None,
            dict_load_func_annotation=None,
            yaml_path: Union[str, None] = None,
            sample_fn: Union[str, None] = None,
            sample_fns: Union[List[str], None] = None,
            additional_annotation_key: Union[str, None] = None,
            cache_metadata: bool = False,
            verbose: int = 0,
            **kwargs
    ):
        super().__init__(
            data_path=data_path,
            meta_path=meta_path,
            cache_path=cache_path,
            load_func=load_func,
            sample_fn=sample_fn,
            sample_fns=sample_fns,
        )
        # General keys are defined in the shared IDs object. Further down, the species specific one is loaded to
        # disambiguate species-dependent differences.
        self._adata_ids_cellxgene = AdataIdsCellxgene()
        self._cache_metadata = cache_metadata
        self._collection = None
        self.collection_id = collection_id
        self.supplier = "cellxgene"
        doi = [x['link_url'] for x in self.collection["links"] if x['link_type'] == 'DOI']
        self.doi_journal = collection_id if len(doi) == 0 else doi[0]  # TODO access journal DOI explicitly.
        self.id = sample_fn
        # Set h5ad download URLs:
        download_url_data = []
        for asset in self._collection_dataset['dataset_assets']:
            if asset['filetype'].lower() == "h5ad":
                download_url_data.append(CELLXGENE_PRODUCTION_ENDPOINT + DOWNLOAD_DATASET + asset['dataset_id'] +
                                         "/asset/" + asset['id'])
        self.download_url_data = download_url_data

        # Set dataset-wise attributes based on object preview from REST API (without h5ad download):
        # Set organism first so that other terms can access this attribute (e.g. developmental_stage ontology).
        reordered_keys = ["organism"] + [x for x in self._adata_ids_cellxgene.dataset_keys if x != "organism"]
        for k in reordered_keys:
            val = self._collection_dataset[getattr(self._adata_ids_cellxgene, k)]
            # Unique label if list is length 1:
            # Otherwise do not set property and resort to cell-wise labels.
            v_clean = clean_cellxgene_meta_uns(k=k, val=val, adata_ids=self._adata_ids_cellxgene)
            # Set as single element or list if multiple entries are given.
            if len(v_clean) > 1:
                v_clean = v_clean[0]
            try:
                setattr(self, k, v_clean)
            except ValueError as e:
                if verbose > 0:
                    print(f"WARNING: {e} in {self.collection_id} and data set {self.id}")

        self._adata_ids_cellxgene = AdataIdsCellxgene_v2_0_0()
        # Add author information.  # TODO need to change this to contributor?
        self.author = "cellxgene"
        self.layer_processed = "X"
        # TODO extract counts vs processed layers
        # The h5ad objects from cellxgene follow a particular structure and the following attributes are guaranteed to
        # be in place. Note that these point at the anndata instance and will only be available for evaluation after
        # download. See below for attributes that are lazily available
        self.cell_type_obs_key = self._adata_ids_cellxgene.cell_type
        self.development_stage_obs_key = self._adata_ids_cellxgene.development_stage
        self.disease_obs_key = self._adata_ids_cellxgene.disease
        self.ethnicity_obs_key = self._adata_ids_cellxgene.ethnicity
        self.sex_obs_key = self._adata_ids_cellxgene.sex
        self.organ_obs_key = self._adata_ids_cellxgene.organ
        self.state_exact_obs_key = self._adata_ids_cellxgene.state_exact

        self.feature_symbol_var_key = self._adata_ids_cellxgene.feature_symbol

        self._unknown_celltype_identifiers = self._adata_ids_cellxgene.unknown_metadata_identifier

    @property
    def _collection_cache_dir(self):
        """
        The cache dir is in a cache directory in the homedirectory of the user by default and can be user modified.
        """
        return settings.cachedir_databases_cellxgene

    @property
    def _collection_cache_fn(self):
        return os.path.join(self._collection_cache_dir, self.collection_id + ".pickle")

    @property
    def collection(self):
        """
        Cached collection meta data.

        Note on caching: updates to the remote collection break these caches.
        Disbale caching are clear caches manually (~/.cache/sfaira/dataset_meta/cellxgene) if this causes issues.
        """
        if self._collection is None:
            # Check if cached:
            if os.path.exists(self._collection_cache_fn) and self._cache_metadata:
                with open(self._collection_cache_fn, "rb") as f:
                    self._collection = pickle.load(f)
            else:
                # Download and cache:
                self._collection = get_collection(collection_id=self.collection_id)
                if self._cache_metadata:
                    with open(self._collection_cache_fn, "wb") as f:
                        pickle.dump(obj=self._collection, file=f)
        return self._collection

    @property
    def _collection_dataset(self):
        return self.collection['datasets'][self._sample_fns.index(self.sample_fn)]

    @property
    def directory_formatted_doi(self) -> str:
        return self.collection_id

    @property
    def doi_cleaned_id(self):
        return self.id

    def load(
            self,
            load_raw: bool = False,
            allow_caching: bool = True,
            set_metadata: bool = True,
            **kwargs
    ):
        # Invoke load with cellxgene adapted parameters:
        #  - Never cache as the cellxgene objects already fast to read.
        super().load(
            load_raw=True,
            allow_caching=False,
            set_metadata=set_metadata,
            adata_ids=self._adata_ids_cellxgene,
            **kwargs
        )

    def download(self, filetype: str = "h5ad", verbose: int = 0):
        """

        Only download if file does not already exist.

        :param filetype: File type to download.

            - "h5ad"
            - "rds"
            - "loom"
        """
        counter = 0
        if not os.path.exists(os.path.join(self.data_dir_base, self.directory_formatted_doi)):
            os.makedirs(os.path.join(self.data_dir_base, self.directory_formatted_doi))
        for asset in self._collection_dataset['dataset_assets']:
            if asset['filetype'].lower() == filetype:
                # Only download if file does not already exist:
                fn = cellxgene_fn(dir=self.data_dir, dataset_id=self.sample_fn)
                if not os.path.isfile(fn):
                    counter += 1
                    assert counter < 2, f"found more than one {filetype} for data set {self.sample_fn}"
                    url = CELLXGENE_PRODUCTION_ENDPOINT + DOWNLOAD_DATASET + asset['dataset_id'] + "/asset/" + \
                        asset['id']
                    r = requests.post(url)
                    r.raise_for_status()
                    presigned_url = r.json()['presigned_url']
                    # Report:
                    headers = {'range': 'bytes=0-0'}
                    r1 = requests.get(presigned_url, headers=headers)
                    if r1.status_code == requests.codes.partial:
                        if verbose > 0:
                            print(f"Downloading {r1.headers['Content-Range']} from {r1.headers['Server']}")
                    # Download:
                    open(fn, 'wb').write(get_data(presigned_url=presigned_url))

    def show_summary(self):
        uuid_session = str(uuid.uuid4())
        display_html('<div id="{}" style="height: 600px; width:100%;"></div>'.format(uuid_session), raw=True)
        display_javascript("""
        require(["https://rawgit.com/caldwell/renderjson/master/renderjson.js"], function() {
          document.getElementById('%s').appendChild(renderjson(%s))
        });
        """ % (uuid_session, json.dumps(self._collection_dataset)), raw=True)


def load(data_dir, sample_fn, adata_ids: AdataIdsCellxgene, **kwargs):
    """
    Generalised load function for cellxgene-provided data sets.

    This function corresponds to the dataset-wise load() functions defined in standard sfaira data loaders.
    """
    fn = cellxgene_fn(dir=data_dir, dataset_id=sample_fn)
    adata = anndata.read_h5ad(fn)
    if adata.raw is not None:  # TODO still need this?
        adata.X = adata.raw.X
        del adata.raw
    for k in adata_ids.ontology_constrained:
        col_name = getattr(adata_ids, k)
        if col_name in adata.obs.columns:
            adata.obs[col_name] = clean_cellxgene_meta_obs(k=k, val=adata.obs[col_name].values, adata_ids=adata_ids)
    return adata



import anndata
import os
import scipy.sparse

from sfaira.data import DatasetBase

SAMPLE_FNS = [
    "madissoon19_lung.processed.h5ad",
    "oesophagus.cellxgene.h5ad",
    "spleen.cellxgene.h5ad",
]


class Dataset(DatasetBase):
    """
    ToDo: patient information in .obs["patient"] and sample information in .obs["sample"] (more samples than patients)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.sample_fn == "madissoon19_lung.processed.h5ad":
            self.download_url_data = "https://covid19.cog.sanger.ac.uk/madissoon19_lung.processed.h5ad"
            self.feature_id_var_key = "gene.ids.HCATisStab7509734"
        elif self.sample_fn == "oesophagus.cellxgene.h5ad":
            self.download_url_data = \
                "https://cellgeni.cog.sanger.ac.uk/tissue-stability/tissue-stability/oesophagus.cellxgene.h5ad"
            # Associated DCP: https://data.humancellatlas.org/explore/projects/c4077b3c-5c98-4d26-a614-246d12c2e5d7
            self.feature_id_var_key = "gene_ids-HCATisStab7413619"
        else:
            self.download_url_data = \
                "https://cellgeni.cog.sanger.ac.uk/tissue-stability/tissue-stability/spleen.cellxgene.h5ad"
            self.feature_id_var_key = "gene_ids-HCATisStab7463846"

        self.download_url_meta = None

        self.assay_sc = "10x 3' v2"
        self.author = "Madissoon"
        self.disease = "healthy"
        self.doi_journal = "10.1186/s13059-019-1906-x"
        self.layer_counts = "X"  # ToDo "madissoon19_lung.processed.h5ad" is close to integer but not quire (~1e-4)
        self.organ = "lung parenchyma" if self.sample_fn == "madissoon19_lung.processed.h5ad" else \
            "esophagus" if self.sample_fn == "oesophagus.cellxgene.h5ad" else "spleen"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.year = 2019
        self.sample_source = "primary_tissue"

        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"
        self.cell_type_obs_key = "Celltypes"

        self.set_dataset_id(idx=1)


def load(data_dir, sample_fn, **kwargs):
    fn = os.path.join(data_dir, sample_fn)
    adata = anndata.read(fn)
    if sample_fn != "madissoon19_lung.processed.h5ad":
        adata.X = adata.X.multiply(scipy.sparse.csc_matrix(adata.obs["n_counts"].values[:, None]))\
            .multiply(1 / 10000)
    # Cell type column called differently in madissoon19_lung.processed.h5ad:
    if sample_fn == "madissoon19_lung.processed.h5ad":
        adata.obs["Celltypes"] = adata.obs["CellType"]
        del adata.obs["CellType"]

    return adata
